# standard library
import os
import sys
import unittest

# utilities
import _compat

# tested class
from cpymad.madx import Madx, CommandLog


class TestMadx(unittest.TestCase, _compat.TestCase):

    """
    Test methods for the Madx class.

    The tests are directly based on the specifics of the sequence in

        test/data/lebt/init.madx

    Please compare this file for reference.
    """

    def setUp(self):
        self.mad = Madx(command_log=CommandLog(sys.stdout, 'X:> '))
        here = os.path.dirname(__file__)
        there = os.path.join(here, 'data', 'lebt', 'init.madx')
        self.doc = open(there).read()
        for line in self.doc.splitlines():
            self.mad._libmadx.input(line)

    def tearDown(self):
        del self.mad

    def test_version(self):
        """Check that the Madx.version attribute can be used as expected."""
        version = self.mad.version
        # check format:
        major, minor, mini = map(int, version.release.split('.'))
        # We need at least MAD-X 5.02.03:
        self.assertGreaterEqual(major, 5)
        self.assertGreaterEqual(minor, 2)
        self.assertGreaterEqual(mini, 3)
        # check format:
        year, month, day = map(int, version.date.split('.'))
        self.assertGreaterEqual(year, 2014)
        self.assertGreaterEqual(month, 1)
        self.assertGreaterEqual(day, 1)
        self.assertLessEqual(month, 12)
        self.assertLessEqual(day, 31)

    def test_independent_instances(self):
        # create a second Madx instance (1st one is created in setUp)
        madxness = Madx()
        # Check independence by defining a variable differently in each
        # instance:
        self.mad.input('ANSWER=42;')
        madxness.input('ANSWER=43;')
        self.assertEqual(self.mad.evaluate('ANSWER'), 42);
        self.assertEqual(madxness.evaluate('ANSWER'), 43);

    def test_command_log(self):
        """Check that the command log contains all input commands."""
        # create a new Madx instance that uses the history feature:
        history_filename = '_test_madx.madx.tmp'
        mad = Madx(command_log=history_filename)
        # feed some input and compare with history file:
        for line in self.doc.splitlines():
            mad.input(line)
        with open(history_filename) as history_file:
            history = history_file.read()
        self.assertEqual(history.strip(), self.doc.strip())
        # remove history file
        del mad
        os.remove(history_filename)

    # TODO:
    # def test_command(self):
    # def test_help(self):
    # def test_call(self):

    def _check_twiss(self, seq_name):
        beam = 'beam, ex=1, ey=2, particle=electron, sequence={0};'.format(seq_name)
        self.mad.command(beam)
        initial = dict(alfx=0.5, alfy=1.5,
                       betx=2.5, bety=3.5)
        # by explicitly specifying the 'columns' parameter a persistent copy
        # is returned. We check that this copy contains the data we want and
        # that it has a 'summary' attribute:
        twiss = self.mad.twiss(sequence=seq_name,
                               columns=['betx', 'bety', 'alfx', 'alfy'],
                               **initial)
        betx, bety = twiss['betx'], twiss['bety']
        alfx, alfy = twiss['alfx'], twiss['alfy']
        # Check initial values:
        self.assertAlmostEqual(twiss['alfx'][0], initial['alfx'])
        self.assertAlmostEqual(twiss['alfy'][0], initial['alfy'])
        self.assertAlmostEqual(twiss['betx'][0], initial['betx'])
        self.assertAlmostEqual(twiss['bety'][0], initial['bety'])
        self.assertAlmostEqual(twiss.summary['ex'], 1)
        self.assertAlmostEqual(twiss.summary['ey'], 2)
        # Check that keys are all lowercase:
        for k in twiss:
            self.assertEqual(k, k.lower())
        for k in twiss.summary:
            self.assertEqual(k, k.lower())

    def test_twiss_1(self):
        self._check_twiss('s1')     # s1 can be computed at start
        self._check_twiss('s1')     # s1 can be computed multiple times
        self._check_twiss('s2')     # s2 can be computed after s1

    def test_twiss_2(self):
        self._check_twiss('s2')     # s2 can be computed at start
        self._check_twiss('s1')     # s1 can be computed after s2

    def test_twiss_with_range(self):
        beam = 'beam, ex=1, ey=2, particle=electron, sequence=s1;'
        self.mad.command(beam)
        params = dict(alfx=0.5, alfy=1.5,
                      betx=2.5, bety=3.5,
                      columns=['betx', 'bety'],
                      sequence='s1')
        # Compute TWISS on full sequence, then on a sub-range, then again on
        # the full sequence. This checks that none of the range selections
        # have side-effects on each other:
        betx_full1 = self.mad.twiss(**params)['betx']
        betx_range = self.mad.twiss(range=('dr[2]', 'sb'), **params)['betx']
        betx_full2 = self.mad.twiss(**params)['betx']
        # Check that the results have the expected lengths:
        self.assertEqual(len(betx_full1), 9)
        self.assertEqual(len(betx_range), 4)
        self.assertEqual(len(betx_full2), 9)
        # Check numeric results. Since the first 3 elements of range and full
        # sequence are identical, equal results are expected. And non-equal
        # results afterwards.
        self.assertAlmostEqual(betx_range[0], betx_full1[1]) # dr:2, dr:1
        self.assertAlmostEqual(betx_range[1], betx_full1[2]) # qp:2, qp:1
        self.assertAlmostEqual(betx_range[2], betx_full1[3]) # dr:3, dr:2
        self.assertNotAlmostEqual(betx_range[3], betx_full1[4]) # sb, qp:2

    # def test_survey(self):
    # def test_aperture(self):
    # def test_use(self):
    # def test_match(self):
    # def test_verbose(self):

    def test_active_sequence(self):
        self.mad.command('beam, ex=1, ey=2, particle=electron, sequence=s1;')
        self.mad.active_sequence = 's1'
        self.assertEqual(self.mad.active_sequence.name, 's1')

    def test_get_sequence(self):
        with self.assertRaises(KeyError):
            self.mad.sequences['sN']
        s1 = self.mad.sequences['s1']
        self.assertEqual(s1.name, 's1')

    def test_get_sequences(self):
        seqs = self.mad.sequences
        self.assertItemsEqual(seqs, ['s1', 's2'])

    def test_evaluate(self):
        val = self.mad.evaluate("1/QP_K1")
        self.assertAlmostEqual(val, 0.5)

    def test_set_value(self):
        self.mad.set_value('FOO', 1)
        self.mad.set_value('BAR', 'FOO')
        self.mad.set_value('FOO', 2)
        self.assertEqual(self.mad.evaluate('FOO'), 2)
        self.assertEqual(self.mad.evaluate('BAR'), 1)

    def test_set_expression(self):
        self.mad.set_expression('FOO', 'BAR')
        self.mad.set_value('BAR', 42)
        self.mad.evaluate('FOO')
        self.assertEqual(self.mad.evaluate('FOO'), 42)
        self.mad.set_value('BAR', 43)
        self.assertEqual(self.mad.evaluate('FOO'), 43)

    # def test_sequence_beam(self):
    # def test_sequence_twiss(self):
    # def test_sequence_twissname(self):

    def _get_elems(self, seq_name):
        elems = self.mad.sequences[seq_name].elements
        elem_dict = dict((el['name'], el) for el in elems)
        elem_idx = dict((el['name'], i) for i, el in enumerate(elems))
        return elem_dict, elem_idx

    def test_sequence_get_elements_s1(self):
        s1, idx = self._get_elems('s1')
        qp1 = s1['qp:1']
        qp2 = s1['qp:2']
        sb1 = s1['sb:1']
        self.assertLess(idx['qp:1'], idx['qp:2'])
        self.assertLess(idx['qp:2'], idx['sb:1'])
        self.assertAlmostEqual(qp1['at'], 1)
        self.assertAlmostEqual(qp2['at'], 3)
        self.assertAlmostEqual(sb1['at'], 5)
        self.assertAlmostEqual(qp1['l'], 1)
        self.assertAlmostEqual(qp2['l'], 1)
        self.assertAlmostEqual(sb1['l'], 2)
        self.assertAlmostEqual(float(qp1['k1']), 2)
        self.assertAlmostEqual(float(qp2['k1']), 2)
        self.assertAlmostEqual(float(sb1['angle']), 3.14/4)
        self.assertEqual(str(qp1['k1']).lower(), "qp_k1")

    def test_sequence_get_elements_s2(self):
        s2, idx = self._get_elems('s2')
        qp1 = s2['qp1:1']
        qp2 = s2['qp2:1']
        self.assertLess(idx['qp1:1'], idx['qp2:1'])
        self.assertAlmostEqual(qp1['at'], 0)
        self.assertAlmostEqual(qp2['at'], 1)
        self.assertAlmostEqual(qp1['l'], 1)
        self.assertAlmostEqual(qp2['l'], 2)
        self.assertAlmostEqual(float(qp1['k1']), 3)
        self.assertAlmostEqual(float(qp2['k1']), 2)

    # def test_sequence_get_expanded_elements(self):

    def test_crash(self):
        """Check that a RuntimeError is raised in case MAD-X crashes."""
        # a.t.m. MAD-X crashes on this input, because the L (length)
        # parametere is missing:
        self.assertRaises(RuntimeError, self.mad.input, 'XXX: sequence;')

    def test_sequence_elements(self):
        elements = self.mad.sequences['s1'].elements
        iqp2 = elements.index('qp:2')
        qp1 = elements['qp:1']
        qp2 = elements[iqp2]
        self.assertAlmostEqual(qp1['at'], 1)
        self.assertAlmostEqual(qp2['at'], 3)
        self.assertEqual(iqp2, elements.at(3.1))

    def test_sequence_expanded_elements(self):
        beam = 'beam, ex=1, ey=2, particle=electron, sequence=s1;'
        self.mad.command(beam)
        self.mad.use('s1')
        elements = self.mad.sequences['s1'].expanded_elements
        iqp2 = elements.index('qp:2')
        qp1 = elements['qp:1']
        qp2 = elements[iqp2]
        self.assertAlmostEqual(qp1['at'], 1)
        self.assertAlmostEqual(qp2['at'], 3)
        self.assertEqual(iqp2, elements.at(3.1))


if __name__ == '__main__':
    unittest.main()



# encoding: utf-8
"""
Resource provider for python package resources.
"""

import pkg_resources
from contextlib import contextmanager, closing
from shutil import rmtree
from os import remove
from os.path import isdir
from io import StringIO, BytesIO, open

from .base import ResourceProvider


__all__ = [
    'PackageResource',
]


_manager = pkg_resources.ResourceManager()


class PackageResource(ResourceProvider):

    """
    Provider for resources that are inside a python package.

    This can be used to access data that resides within .egg files as well
    as data that is accessible through the filesystem.

    Uses pkg_resources.resource_stream() to open resources and
    pkg_resources.listdir() to list available resources.
    """

    def __init__(self, package, path=''):
        """
        Initialize package resource provider.

        :param string package: python package/module name or object
        :param string path: name of a resource relative to the package
        """
        self.package = package
        self.path = path
        self._manager = _manager
        self._provider = pkg_resources.get_provider(package)

    def open(self, name='', encoding=None):
        # The python2 implementation of ZipProvider.get_resource_stream
        # uses cStringIO.StringIO which is unacceptable for our purposes.
        # It is not usable as context manager, it fails to support unicode,
        # it cannot be wrapped by TextIOWrapper. This is how it should have
        # been done from the beginning:
        if self._is_filesystem:
            filename = self._provider.get_resource_filename(
                self._manager,
                self._get_path(name))
            if encoding:
                return open(filename, 'rt', encoding=encoding)
            else:
                return open(filename, 'rb')
        else:
            data = self.load(name, encoding)
            if encoding:
                return StringIO(data)
            else:
                return BytesIO(data)

    def load(self, name='', encoding=None):
        data = self._provider.get_resource_string(
                self._manager,
                self._get_path(name))
        if encoding:
            return data.decode(encoding)
        else:
            return data

    def listdir(self, name=''):
        return self._provider.resource_listdir(self._get_path(name))

    def get(self, name=''):
        return self.__class__(self.package, self._get_path(name))

    @contextmanager
    def filename(self, name=''):
        filename = self._provider.get_resource_filename(
            self._manager,
            self._get_path(name))
        try:
            yield filename
        finally:
            if self._is_extracted:
                # there is also pkg_resources.cleanup_resources but this
                # deletes all cached resources. Furthermore, that method
                # seems to be a NOOP, currently.
                if isdir(filename):
                    rmtree(filename)
                else:
                    remove(filename)


    def _get_path(self, name):
        if not name:
            return self.path
        elif isinstance(name, list):
            return '/'.join([self.path] + name)
        else:
            return '/'.join([self.path] + name.split('/'))

    @property
    def _is_filesystem(self):
        return isinstance(self._provider, pkg_resources.DefaultProvider)

    @property
    def _is_extracted(self):
        # This check is not very forward compatible, but its the best I got:
        return isinstance(self._provider, pkg_resources.ZipProvider)

    def provider(self):
        parts = self.path.rsplit('/', 1)
        return self.__class__(self.package, parts[0] if len(parts) > 1 else '')



"""
Models encapsulate metadata for accelerator machines.

For more information about models, see :class:`Model`.
"""

from __future__ import absolute_import

import logging
import os

from .madx import Madx
from .util import is_match_param
from .resource.file import FileResource


__all__ = [
    'Model',
    'Beam',
    'Optic',
    'Sequence',
    'Range',
    'Locator',
]


def _deserialize(data, cls, *args):
    """Create an instance dictionary from a data dictionary."""
    return {key: cls(key, val, *args) for key, val in data.items()}


def _serialize(data):
    """Create a data dictionary from an instance dictionary."""
    return {key: val.data for key, val in data.items()}


class Model(object):

    """
    A model is a complete description of an accelerator machine.

    This class is used to bundle all metadata related to an accelerator and
    all its configurations. It takes care of loading the proper MAD-X files
    when needed. Models are conceptually derived from the JMad models, but
    have evolved to a more pythonic and simple API.

    To create a model instance from a model definition file, use the
    ``Model.load`` constructor.

    Only GET access is allowed to all instance variables at the moment.

    Public attributes:

    :ivar str name: model name
    :ivar dict beams: known :class:`Beam` objects
    :ivar dict optics: known :class:`Optic` objects
    :ivar dict sequences: known :class:`Sequence` objects
    :ivar Madx madx: handle to the MAD-X library

    Private variables:

    :ivar dict _data: model definition data
    :ivar ResourceProvider _repo: resource access

    The following example demonstrates the basic usage:

    .. code-block:: python

        from cpymad.model import Model

        model = Model.load('/path/to/model/definition.cpymad.yml')

        twiss = model.default_sequence.twiss()

        print("max/min beta x:", max(twiss['betx']), min(twiss['betx']))
        print("ex: {0}, ey: {1}", twiss.summary['ex'], twiss.summary['ey'])
    """

    # current version of model API
    API_VERSION = 0

    def __init__(self, data, repo, madx):
        """
        Initialize a Model object.

        :param dict data: model definition data
        :param ResourceProvider repo: resource repository
        :param Madx madx: MAD-X instance to use
        """
        self.check_compatibility(data)
        # init instance variables
        self._data = data
        self._repo = repo
        self.madx = madx
        self._loaded = False
        # create Beam/Optic/Sequence instances:
        self.beams = _deserialize(data['beams'], Beam, self)
        self.optics = _deserialize(data['optics'], Optic, self)
        self.sequences = _deserialize(data['sequences'], Sequence, self)

    @classmethod
    def check_compatibility(cls, data):
        """
        Check a model definition for compatibility.

        :param dict data: a model definition to be tested
        :raises ValueError: if the model definition is incompatible
        """
        model_api = data.get('api_version', 'undefined')
        if model_api != cls.API_VERSION:
            raise ValueError(("Incompatible model API version: {!r},\n"
                              "              Required version: {!r}")
                             .format(model_api, cls.API_VERSION))

    @classmethod
    def load(cls,
             name,
             # *,
             # These should be passed as keyword-only parameters:
             locator=None,
             madx=None,
             command_log=None,
             error_log=None):
        """
        Create Model instance from a model definition file.

        :param str name: model definition file name
        :param Locator locator: model locator
        :param Madx madx: MAD-X instance to use
        :param str command_log: history file name; use only if madx is None!
        :param logging.Logger error_log:

        If the ``locator`` is not specified ``name`` is assumed to be an
        absolute path of a model definition file living in the ordinary file
        system.
        """
        if locator is None:
            path, name = os.path.split(name)
            locator = Locator(FileResource(path))
        data = locator.get_definition(name)
        repo = locator.get_repository(data)
        if madx is None:
            if error_log is None:
                error_log = logging.getLogger(__name__ + '.' + name)
            madx = Madx(command_log=command_log, error_log=error_log)
            madx.verbose(False)
        elif command_log is not None:
            raise ValueError("'command_log' cannot be used with 'madx'")
        elif error_log is not None:
            raise ValueError("'error_log' cannot be used with 'madx'")
        model = cls(data, repo=repo, madx=madx)
        model.init()
        return model

    def init(self):
        """Load model in MAD-X interpreter."""
        if self._loaded:
            return
        self._loaded = True
        self._load(*self._data['init-files'])

    def __repr__(self):
        return "{0}({1!r})".format(self.__class__.__name__, self.name)

    @property
    def name(self):
        """Model name."""
        return self._data['name']

    @property
    def data(self):
        """Get a serializable representation of this model."""
        data = self._data.copy()
        data['beams'] = _serialize(self.beams)
        data['optics'] = _serialize(self.optics)
        data['sequences'] = _serialize(self.sequences)
        return data

    @property
    def default_optic(self):
        """Get default Optic."""
        return self.optics[self._data['default-optic']]

    @property
    def default_sequence(self):
        """Get default Sequence."""
        return self.sequences[self._data['default-sequence']]

    # TODO: add setters for default_optic / default_sequence
    # TODO: remove default_sequence?

    def _load(self, *files):
        """Load MAD-X files in interpreter."""
        for file in files:
            with self._repo.get(file).filename() as fpath:
                self.madx.call(fpath)


class Beam(object):

    """
    Beam for :class:`Model`.

    A beam defines the mass, charge, energy, etc. of the particles moved
    through the accelerator.

    Public attributes:

    :ivar str name: beam name
    :ivar dict data: beam parameters (keywords to BEAM command in MAD-X)

    Private variables:

    :ivar Model _model: owning model
    :ivar bool _loaded: beam has been initialized in MAD-X
    """

    def __init__(self, name, data, model):
        """Initialize instance variables."""
        self.name = name
        self.data = data
        self._model = model
        self._loaded = False

    def init(self):
        """Define the beam in MAD-X."""
        if self._loaded:
            return
        self._loaded = True
        self._model.init()
        self._model.madx.command.beam(**self.data)


class Optic(object):

    """
    Optic for :class:`Model`.

    An optic (as far as I understand) defines a variant of the accelerator
    setup, e.g. different injection mechanisms.

    Public attributes:

    :ivar str name: optic name

    Private variables:

    :ivar dict _data: optic definition
    :ivar Model _model: owning model
    :ivar bool _loaded: beam has been initialized in MAD-X
    """

    def __init__(self, name, data, model):
        """Initialize instance variables."""
        self.name = name
        self.data = data
        self._model = model
        self._loaded = False

    def init(self):
        """Load the optic in the MAD-X process."""
        if self._loaded:
            return
        self._loaded = True
        self._model.init()
        self._model._load(*self.data.get('init-files', ()))


class Sequence(object):

    """
    Sequence for :class:`Model`.

    A sequence defines an arrangement of beam line elements. It can be
    subdivided into multiple ranges.

    Public attributes:

    :ivar str name: sequence name
    :ivar dict ranges: known :class:`Range` objects

    Private variables:

    :ivar dict _data:
    :ivar Model _model:
    """

    def __init__(self, name, data, model):
        """Initialize instance variables."""
        self.name = name
        self._data = data
        self._model = model
        self.ranges = _deserialize(data['ranges'], Range, self)

    def init(self):
        """Load model in MAD-X interpreter."""
        self._model.init()
        self.beam.init()

    @property
    def data(self):
        """Get a serializable representation of this sequence."""
        data = self._data.copy()
        data['ranges'] = _serialize(self.ranges)
        return data

    @property
    def beam(self):
        """Get :class:`Beam` instance for this sequence."""
        return self._model.beams[self._data['beam']]

    @property
    def default_range(self):
        """Get default :class:`Range`."""
        return self.ranges[self._data['default-range']]

    def range(self, start, stop):
        """Create a :class:`Range` within (start, stop) for this sequence."""
        # TODO
        raise NotImplementedError()

    # MAD-X commands:

    def twiss(self, **kwargs):
        """Execute a TWISS command on the default range."""
        return self.default_range.twiss(**kwargs)

    def survey(self, **kwargs):
        """Run SURVEY on this sequence."""
        self.init()
        return self._model.madx.survey(sequence=self.name, **kwargs)

    def match(self, **kwargs):
        """Run MATCH on this sequence."""
        return self.default_range.match(**kwargs)


class Range(object):

    """
    Range for :class:`Model`.

    A range is a subsequence of elements within a :class:`Sequence`.

    Public attributes:

    :ivar str name: sequence name

    Private variables:

    :ivar dict _data:
    :ivar Sequence _sequence:
    """

    def __init__(self, name, data, sequence):
        """Initialize instance variables."""
        self.name = name
        self.data = data
        self._sequence = sequence

    def init(self):
        """Load model in MAD-X interpreter."""
        self._sequence.init()

    @property
    def bounds(self):
        """Get a tuple (first, last)."""
        return (self.data["madx-range"]["first"],
                self.data["madx-range"]["last"])

    @property
    def offsets_file(self):
        """Get a :class:`ResourceProvider` for the offsets file."""
        if 'aper-offset' not in self.data:
            return None
        repo = self._sequence._model._repo
        return _repo.get(self.data['aper-offset'])

    def twiss(self, **kwargs):
        """Run TWISS on this range."""
        self.init()
        kw = self._set_twiss_init(kwargs)
        madx = self._sequence._model.madx
        result = madx.twiss(sequence=self._sequence.name,
                            range=self.bounds, **kw)
        return result

    def match(self, **kwargs):
        """Perform a MATCH operation on this range."""
        self.init()
        kw = self._set_twiss_init(kwargs)
        kw['twiss_init'] = {
            key: val
            for key, val in kw['twiss_init'].items()
            if is_match_param(key)
        }
        madx = self._sequence._model.madx
        return madx.match(sequence=self._sequence.name,
                          range=self.bounds, **kw)

    @property
    def initial_conditions(self):
        """
        Return a dict of all defined initial conditions.

        Each item is a dict of TWISS parameters.
        """
        return self.data['twiss-initial-conditions']

    @property
    def default_initial_conditions(self):
        """Return the default twiss initial conditions."""
        return self.initial_conditions[self.data['default-twiss']]

    def _set_twiss_init(self, kwargs):
        kw = kwargs.copy()
        twiss_init = kw.get('twiss_init', {}).copy()
        twiss_init.update(self.default_initial_conditions)
        kw['twiss_init'] = twiss_init
        return kw


class Locator(object):

    """
    Model locator for yaml files that contain multiple model definitions.

    These are the model definition files that are currently used by default
    for filesystem resources.

    Serves the purpose of locating models and returning corresponding
    resource providers.
    """

    ext = '.cpymad.yml'

    def __init__(self, resource_provider):
        """
        Initialize a merged model locator instance.

        The resource_provider parameter must be a ResourceProvider instance
        that points to the filesystem location where the .cpymad.yml model
        files are stored.
        """
        self._repo = resource_provider

    def list_models(self, encoding='utf-8'):
        """
        Iterate all available models.

        Returns an iterable that may be a generator object.
        """
        for res_name in self._repo.listdir_filter(ext=self.ext):
            yield res_name[:-len(self.ext)]

    def get_definition(self, name, encoding='utf-8'):
        """
        Get the first found model with the specified name.

        :returns: the model definition
        :raises ValueError: if no model with the given name is found.
        """
        try:
            if not name.endswith(self.ext):
                name += self.ext
            return self._repo.yaml(name, encoding=encoding)
        except IOError:
            raise ValueError("The model {!r} does not exist in the database"
                             .format(name))

    def get_repository(self, data):
        """
        Get the resource loader for the given model.
        """
        # instantiate the resource providers for model resource data
        return self._repo.get(data['path-offset'])



"""
cpymad.libmadx is a Cython binding for the MAD-X library.

This file contains declarations of data structures and exported functions
of the C-API of MAD-X.
"""

# Data structures:

# The following declarations are only used by the Cython compiler (so it
# knows what fields exist in the data types). The C compiler will use the
# real definitions in the header files.
# For greater clarity, we should therefore list only those fields, that are
# actually used in the Cython code and add further fields when needed
# later:

cdef extern from "madX/mad_def.h":
    enum:
        NAME_L

cdef extern from "madX/mad_gcst.h":
    # NOTE: C API uses "const char*"
    char* version_name
    char* version_date

cdef extern from "madX/mad_array.h":
    struct char_array:
        int curr
        char* c

    struct char_p_array:
        int curr
        char** p

    struct int_array:
        int curr
        int* i

    struct double_array:
        int curr
        double* a

cdef extern from "madX/mad_name.h":
    struct name_list:
        int curr
        int* inform
        char** names

cdef extern from "madX/mad_table.h":
    struct table:
        char[NAME_L] name
        int curr
        char_p_array* header
        char*** s_cols
        double** d_cols
        name_list* columns

    struct table_list:
        int curr
        name_list* names
        table** tables

    struct column_info:
        void* data
        int length
        char datatype
        char datasize

cdef extern from "madX/mad_expr.h":
    struct expression:
        char* string
        double value

    struct expr_list:
        int curr
        expression** list

cdef extern from "madX/mad_cmdpar.h":
    struct command_parameter:
        char[NAME_L] name
        int type
        int c_type
        double double_value
        double c_min
        double c_max
        expression* expr
        expression* min_expr
        expression* max_expr
        char* string
        int stamp
        double_array* double_array
        expr_list* expr_list
        char_p_array* m_string

    struct command_parameter_list:
        int curr
        command_parameter** parameters

cdef enum:
    PARAM_TYPE_LOGICAL = 0
    PARAM_TYPE_INTEGER = 1
    PARAM_TYPE_DOUBLE = 2
    PARAM_TYPE_STRING = 3
    PARAM_TYPE_CONSTRAINT = 4
    PARAM_TYPE_LOGICAL_ARRAY = 10   # I invented this one for symmetry.
    PARAM_TYPE_INTEGER_ARRAY = 11
    PARAM_TYPE_DOUBLE_ARRAY = 12
    PARAM_TYPE_STRING_ARRAY = 13

cdef enum:
    CONSTR_TYPE_MIN = 1
    CONSTR_TYPE_MAX = 2
    CONSTR_TYPE_BOTH = 3
    CONSTR_TYPE_VALUE = 4

cdef extern from "madX/mad_cmd.h":
    struct command:
        int beam_def
        command_parameter_list* par

cdef extern from "madX/mad_elem.h":
    struct element:
        char[NAME_L] name
        command* def_ "def"

cdef extern from "madX/mad_node.h":
    struct node:
        char[NAME_L] name
        char* base_name
        double at_value
        double length
        element* p_elem

    struct node_list:
        int curr
        node** nodes
        name_list* list

cdef extern from "madX/mad_seq.h":
    struct sequence:
        # original sequence
        char[NAME_L] name
        int ref_flag
        node_list* nodes
        command* beam
        # expanded sequence
        int n_nodes
        node** all_nodes
        table* tw_table
        int tw_valid

    struct sequence_list:
        int curr
        name_list* list
        sequence** sequs

cdef enum:
    REF_EXIT = -1
    REF_CENTER = 0
    REF_ENTRY = 1


# Global variables:
cdef extern from "madX/mad_gvar.h":
    sequence* current_sequ      # active sequence
    table_list* table_register  # list of all tables
    char_p_array* tmp_p_array   # temporary buffer for splits
    char_array* c_dum           # another temporary buffer


# Function declarations:
cdef extern from "madX/mad_api.h":
    sequence_list *madextern_get_sequence_list()

cdef extern from "madX/mad_core.h":
    void madx_start()
    void madx_finish()

cdef extern from "madX/mad_name.h":
    int name_list_pos(char*, name_list*)  # NOTE: C API uses "const char*"

cdef extern from "madX/mad_str.h":
    void stolower_nq(char*)
    int mysplit(char*, char_p_array*)

cdef extern from "madX/mad_eval.h":
    void pro_input(char*)

cdef extern from "madX/mad_expr.h":
    expression* make_expression(int, char**)
    double expression_value(expression*, int)
    expression* delete_expression(expression*)

cdef extern from "madX/mad_parse.h":
    void pre_split(char*, char_array*, int)

cdef extern from "madX/mad_table.h":
    char_p_array* table_get_header(char* table_name)
    int table_exists(char* table_name)


# I have no clue why, but for some reason, it is necessary to include
# 'madx.h' (or one of the file it includes). Otherwise, importing the Cython
# module will result in an ImportError: undefined symbol: pro_input
cdef extern from "madX/madx.h":
    pass



# encoding: utf-8
"""
Resource provider for plain filesystem resources.
"""

import os
from io import open
from contextlib import contextmanager

from .base import ResourceProvider


__all__ = [
    'FileResource',
]


class FileResource(ResourceProvider):

    """
    File system resource provider.

    Uses the builtins open() to open ordinary files and os.listdir() to list
    directory contents.
    """

    def __init__(self, path):
        """
        Initialize the filesystem resource provider.

        :param string path: name of a filesystem object (file/folder).
        """
        self.path = path

    def open(self, name='', encoding=None):
        if encoding is None:
            return open(self._get_path(name), 'rb')
        else:
            return open(self._get_path(name), 'rt', encoding=encoding)

    def listdir(self, name=''):
        return os.listdir(self._get_path(name))

    def get(self, name=''):
        return self.__class__(self._get_path(name))

    @contextmanager
    def filename(self, name=''):
        yield self._get_path(name)

    def _get_path(self, name):
        if not name:
            return self.path
        elif isinstance(name, list):
            return os.path.join(self.path, *name)
        else:
            return os.path.join(self.path, *name.split('/'))

    def provider(self):
        return self.__class__(os.path.dirname(self.path))



# encoding: utf-8
"""
Tests for the model.Locator functionality.
"""

# tested classes
from cpymad.model import Locator
from cpymad.resource.file import FileResource

# test utilities
import unittest
import os

__all__ = [
    'TestLocator',
]


class TestLocator(unittest.TestCase):

    """
    Test the Locator (standard filesystem locator).
    """

    unicode_data = u"€®æة《±∓"

    def setUp(self):
        here = os.path.dirname(__file__)
        self.base = os.path.join(here, 'data')
        self.locator = Locator(FileResource(self.base))

    def test_list_models(self):
        """Verify the results of the list_models method."""
        self.assertEqual(
            set(self.locator.list_models()),
            set(['lebt']))

    def test_get(self):
        """Test that ValueError is raised iff model does not exist."""
        # cannot get non-existing model:
        self.assertRaises(ValueError, self.locator.get_definition, 'f')
        # can get existing real models:
        self.locator.get_definition('lebt')

    def test_encoding(self):
        """Test that the model is loaded with the correct encoding."""
        b = self.locator.get_definition('lebt')
        self.assertEqual(b['unicode'],
                         self.unicode_data)

    def test_path_resolution(self):
        """Test the .get and .get_by_dict methods of ModelData."""
        mdef = self.locator.get_definition('lebt')
        files = mdef['files']
        repo = self.locator.get_repository(mdef)
        self.assertEqual(repo.get(files[0]).yaml()['path'], 'lebt/misc.yml')


if __name__ == '__main__':
    unittest.main()



"""
This module defines a convenience layer to access the MAD-X interpreter.

The most interesting class for users is :class:`Madx`.
"""

from __future__ import absolute_import

from functools import partial
import logging
import os
import collections

from . import _rpc
from . import util

try:
    basestring
except NameError:
    basestring = str


class Version(object):

    """Version information struct. """

    def __init__(self, release, date):
        """Store version information."""
        self.release = release
        self.date = date

    def __repr__(self):
        """Show nice version string to user."""
        return "MAD-X {} ({})".format(self.release, self.date)


class ChangeDirectory(object):

    """Context manager for temporarily changing current working directory."""

    def __init__(self, path, _os=os):
        """
        Change the path using the given os module.

        :param str path: new path name
        :param module _os: module with 'getcwd' and 'chdir' functions
        """
        self._os = _os
        # Contrary to common implementations of a similar context manager,
        # we change the path immediately in the constructor. That enables
        # this utility to be used without any 'with' statement:
        if path:
            self._restore = _os.getcwd()
            _os.chdir(path)
        else:
            self._restore = None

    def __enter__(self):
        """Enter 'with' context."""
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit 'with' context and restore old path."""
        if self._restore:
            self._os.chdir(self._restore)


class MadxCommands(object):

    """
    Generic MAD-X command wrapper.

    Raw python interface to issue MAD-X commands. Usage example:

    >>> command = MadxCommands(libmadx.input)
    >>> command('twiss', sequence='LEBT')
    >>> command.title('A meaningful phrase')

    :ivar __dispatch: callable that takes a MAD-X command string.
    """

    def __init__(self, dispatch):
        """Set :ivar:`__dispatch` from :var:`dispatch`."""
        self.__dispatch = dispatch

    def __call__(self, *args, **kwargs):
        """Create and dispatch a MAD-X command string."""
        self.__dispatch(util.mad_command(*args, **kwargs))

    def __getattr__(self, name):
        """Return a dispatcher for a specific command."""
        return partial(self.__call__, name)


def NOP(s):
    """Do nothing."""
    pass


class CommandLog(object):

    """Log MAD-X command history to a file text."""

    @classmethod
    def create(cls, filename):
        """Create CommandLog from filename (overwrite/create)."""
        return cls(open(filename, 'wt'))

    def __init__(self, file, prefix='', suffix='\n'):
        """Create CommandLog from file instance."""
        self._file = file
        self._prefix = prefix
        self._suffix = suffix

    def __call__(self, command):
        """Log a single history line and flush to file immediately."""
        self._file.write(self._prefix + command + self._suffix)
        self._file.flush()


class Madx(object):

    """
    Each instance controls one MAD-X process.

    This class aims to expose a pythonic interface to the full functionality
    of the MAD-X library. For example, when you call the ``twiss`` method, you
    get numpy arrays containing the information from the table generated.
    Furthermore, we try to reduce the amount of commands needed by combining
    e.g. USE, SELECT, and TWISS into the ``twiss`` method itself, and define
    reasonable default patterns/columns.

    The following very simple example demonstrates basic usage:

    .. code-block:: python

        from cpymad.madx import Madx

        m = Madx()

        m.call('my-sequences.seq')
        m.call('my-strengths.str')

        m.command.beam(sequence='myseq1', particle='PROTON')
        # you can also just put an arbitrary MAD-X command string here:
        m.command('beam, sequence=myseq1, particle=PROTON')

        twiss = m.twiss('myseq1')

        # now do your own analysis:
        from matplotlib import pyplot as plt
        plt.plot(twiss['s'], twiss['betx'])
        plt.show()

    By default :class:`Madx` uses a subprocess to execute MAD-X library calls
    remotely via a simple RPC protocol which is defined in :mod:`_rpc`. If
    required this behaviour can be customized by passing a custom ``libmadx``
    object to the constructor. This object must expose an interface similar to
    :mod:`libmadx`.
    """

    def __init__(self, libmadx=None, command_log=None, error_log=None):
        """
        Initialize instance variables.

        :param libmadx: :mod:`libmadx` compatible object
        :param command_log: logs MAD-X history either filename or CommandLog
        :param error_log: logger instance ``logging.Logger``
        """
        # get logger
        if error_log is None:
            error_log = logging.getLogger(__name__)
        # open history file
        if isinstance(command_log, basestring):
            command_log = CommandLog.create(command_log)
        elif command_log is None:
            command_log = NOP
        # start libmadx subprocess
        if libmadx is None:
            svc, proc = _rpc.LibMadxClient.spawn_subprocess()
            libmadx = svc.libmadx
        if not libmadx.is_started():
            libmadx.start()
        # init instance variables:
        self._libmadx = libmadx
        self._command_log = command_log
        self._error_log = error_log

    def __bool__(self):
        """Check if MAD-X is up and running."""
        try:
            return self._libmadx.is_started()
        except (_rpc.RemoteProcessClosed, _rpc.RemoteProcessCrashed):
            return False

    __nonzero__ = __bool__      # alias for python2 compatibility

    @property
    def version(self):
        """Get the MAD-X version."""
        return Version(self._libmadx.get_version_number(),
                       self._libmadx.get_version_date())

    @property
    def command(self):
        """
        Perform a single MAD-X command.

        :param str cmd: command name
        :param kwargs: command parameters
        """
        return MadxCommands(self.input)

    def input(self, text):
        """
        Run any textual MAD-X input.

        :param str text: command text
        """
        # write to history before performing the input, so if MAD-X
        # crashes, it is easier to see, where it happened:
        self._command_log(text)
        try:
            self._libmadx.input(text)
        except _rpc.RemoteProcessCrashed:
            # catch + reraise in order to shorten stack trace (~3-5 levels):
            raise RuntimeError("MAD-X has stopped working!")

    def set_value(self, name, value):
        """
        Set a variable value ("=" operator in MAD-X).

        Example:

            >>> madx.set_value('R1QS1->K1', '42')
            >>> madx.evaluate('R1QS1->K1')
            42
        """
        self.input(name + ' = ' + str(value) + ';')

    def set_expression(self, name, expr):
        """
        Set a variable expression (":=" operator in MAD-X).

        Example:

            >>> madx.set_expression('FOO', 'BAR')
            >>> madx.set_value('BAR', 42)
            >>> madx.evaluate('FOO')
            42
            >>> madx.set_value('BAR', 43)
            >>> madx.evaluate('FOO')
            43
        """
        self.input(name + ' := ' + str(expr) + ';')

    def help(self, cmd=None):
        """
        Show help about a command or list all MAD-X commands.

        :param str cmd: command name
        """
        # The case 'cmd == None' will be handled by mad_command
        # appropriately.
        self.command.help(cmd=cmd)

    def chdir(self, path):
        """
        Change the directory. Can be used as context manager.

        :param str path: new path name
        :returns: a context manager that can change the directory back
        :rtype: ChangeDirectory
        """
        # Note, that the libmadx module includes the functions 'getcwd' and
        # 'chdir' so it can be used as a valid 'os' module for the purposes
        # of ChangeDirectory:
        return ChangeDirectory(path, self._libmadx)

    def call(self, filename, chdir=False):
        """
        CALL a file in the MAD-X interpreter.

        :param str filename: file name with path
        :param bool chdir: temporarily change directory in MAD-X process
        """
        if chdir:
            dirname, basename = os.path.split(filename)
            with self.chdir(dirname):
                self.command.call(file=basename)
        else:
            self.command.call(file=filename)

    def select(self, flag, columns, pattern=[]):
        """
        Run SELECT command.

        :param str flag: one of: twiss, makethin, error, seqedit
        :param list columns: column names
        :param list pattern: selected patterns
        """
        select = self.command.select
        select(flag=flag, clear=True)
        if columns:
            select(flag=flag, column=columns)
        for p in pattern:
            select(flag=flag, pattern=p)

    def twiss(self,
              sequence=None,
              range=None,
              # *,
              # These should be passed as keyword-only parameters:
              twiss_init={},
              columns=None,
              pattern=['full'],
              **kwargs):
        """
        Run SELECT+USE+TWISS.

        :param str sequence: name of sequence
        :param list pattern: pattern to include in table
        :param list columns: columns to include in table, (may be a str)
        :param dict twiss_init: dictionary of twiss initialization variables
        :param bool chrom: Also calculate chromatic functions (slower)
        :param kwargs: further keyword arguments for the MAD-X command

        Note that the kwargs overwrite any arguments in twiss_init.
        """
        self.select('twiss', columns=columns, pattern=pattern)
        sequence = self._use(sequence)
        twiss_init = dict((k, v) for k,v in twiss_init.items()
                          if k not in ['name','closed-orbit'])
        # explicitly specified keyword arguments overwrite values in
        # twiss_init:
        twiss_init.update(kwargs)
        self.command.twiss(sequence=sequence,
                           range=range,
                           **twiss_init)
        return self.get_table('twiss', columns)

    def survey(self,
               sequence=None,
               columns=None,
               pattern=['full'],
               **kwargs):
        """
        Run SELECT+USE+SURVEY.

        :param str sequence: name of sequence
        :param list pattern: pattern to include in table
        :param list columns: Columns to include in table
        :param kwargs: further keyword arguments for the MAD-X command
        """
        self.select('survey', pattern=pattern, columns=columns)
        self._use(sequence)
        self.command.survey(**kwargs)
        return self.get_table('survey', columns)

    def use(self, sequence):
        """
        Run USE to expand a sequence.

        :param str sequence: sequence name
        :returns: name of active sequence
        """
        self.command.use(sequence=sequence)

    def _use(self, sequence):
        """
        USE sequence if it is not active.

        :param str sequence: sequence name, may be None
        :returns: new active sequence name
        :rtype: str
        :raises RuntimeError: if there is no active sequence
        """
        try:
            active_sequence = self.active_sequence
        except RuntimeError:
            if not sequence:
                raise
            active_sequence = None
        else:
            if not sequence:
                sequence = active_sequence
        if (sequence != active_sequence
                or not self._libmadx.is_sequence_expanded(sequence)):
            self.use(sequence)
        return sequence

    def match(self,
              sequence=None,
              constraints=[],
              vary=[],
              weight=None,
              method=('lmdif', {}),
              knobfile=None,
              twiss_init={},
              **kwargs):
        """
        Perform a simple MATCH operation.

        For more advanced cases, you should issue the commands manually.

        :param str sequence: name of sequence
        :param list constraints: constraints to pose during matching
        :param list vary: knob names to be varied
        :param dict weight: weights for matching parameters
        :returns: final knob values
        :rtype: dict

        Example:

        >>> mad.match(
        ...     constraints=[
        ...         dict(range='marker1->betx',
        ...              betx=Constraint(min=1, max=3),
        ...              bety=2)
        ...     ],
        ...     vary=['qp1->k1',
        ...           'qp2->k1'])
        """
        sequence = self._use(sequence)
        twiss_init = dict((k, v) for k,v in twiss_init.items()
                          if k not in ['name','closed-orbit'])
        # explicitly specified keyword arguments overwrite values in
        # twiss_init:
        twiss_init.update(kwargs)

        command = self.command
        # MATCH (=start)
        command.match(sequence=sequence, **twiss_init)
        for c in constraints:
            command.constraint(**c)
        for v in vary:
            command.vary(name=v)
        if weight:
            command.weight(**weight)
        command(method[0], **method[1])
        command.endmatch(knobfile=knobfile)
        return dict((knob, self.evaluate(knob)) for knob in vary)

    def verbose(self, switch=True):
        """Turn verbose output on/off."""
        self.command.option(echo=switch, warn=switch, info=switch)

    def get_table(self, table, columns=None):
        """
        Get the specified table from MAD-X.

        :param str table: table name
        :returns: a proxy for the table data
        :rtype: TableProxy
        """
        proxy = TableProxy(table, self._libmadx)
        if columns is None:
            return proxy
        else:
            return proxy.copy(columns)

    @property
    def active_sequence(self):
        """The active :class:`Sequence` (may be None)."""
        try:
            return Sequence(self._libmadx.get_active_sequence_name(),
                            self._libmadx,
                            _check=False)
        except RuntimeError:
            return None

    @active_sequence.setter
    def active_sequence(self, sequence):
        if isinstance(sequence, Sequence):
            name = sequence.name
        elif isinstance(sequence, basestring):
            name = sequence
        try:
            active_sequence = self.active_sequence
        except RuntimeError:
            self.use(name)
        else:
            if active_sequence.name != name:
                self.use(name)

    def evaluate(self, cmd):
        """
        Evaluates an expression and returns the result as double.

        :param str cmd: expression to evaluate.
        :returns: numeric value of the expression
        :rtype: float
        """
        return self._libmadx.evaluate(cmd)

    @property
    def sequences(self):
        """A dict like view of all sequences in memory."""
        return SequenceMap(self._libmadx)

    @property
    def tables(self):
        """A dict like view of all tables in memory."""
        return TableMap(self._libmadx)


def _map_repr(self):
    """String representation of a custom mapping object."""
    return str(dict(self))


class SequenceMap(collections.Mapping):

    """
    A dict like view of all sequences (:class:`Sequence`) in memory.
    """

    def __init__(self, libmadx):
        self._libmadx = libmadx

    __repr__ = _map_repr
    __str__ = _map_repr

    def __iter__(self):
        return iter(self._libmadx.get_sequence_names())

    def __getitem__(self, name):
        try:
            return Sequence(name, self._libmadx)
        except ValueError:
            raise KeyError

    def __contains__(self, name):
        return self._libmadx.sequence_exists(name)

    def __len__(self):
        return self._libmadx.get_sequence_count()


class TableMap(collections.Mapping):

    """
    A dict like view of all tables (:class:`TableProxy`) in memory.
    """

    def __init__(self, libmadx):
        self._libmadx = libmadx

    __repr__ = _map_repr
    __str__ = _map_repr

    def __iter__(self):
        return iter(self._libmadx.get_table_names())

    def __getitem__(self, name):
        try:
            return TableProxy(name, self._libmadx)
        except ValueError:
            raise KeyError

    def __contains__(self, name):
        return self._libmadx.table_exists(name)

    def __len__(self):
        return self._libmadx.get_table_count()


class Sequence(object):

    """
    MAD-X sequence representation.
    """

    def __init__(self, name, libmadx, _check=True):
        """Store sequence name."""
        self._name = name
        self._libmadx = libmadx
        if _check and not libmadx.sequence_exists(name):
            raise ValueError("Invalid sequence: {!r}".format(name))

    def __str__(self):
        """String representation."""
        return "<{}: {}>".format(self.__class__.__name__, self._name)

    __repr__ = __str__

    @property
    def name(self):
        """Get the name of the sequence."""
        return self._name

    @property
    def beam(self):
        """Get the beam dictionary associated to the sequence."""
        return self._libmadx.get_sequence_beam(self._name)

    @property
    def twiss_table(self):
        """Get the TWISS results from the last calculation."""
        return TableProxy(self.twiss_table_name, self._libmadx)

    @property
    def twiss_table_name(self):
        """Get the name of the table with the TWISS results."""
        return self._libmadx.get_sequence_twiss_table_name(self._name)

    @property
    def elements(self):
        """Get list of elements."""
        return ElementList(self._name, self._libmadx, expanded=False)

    @property
    def expanded_elements(self):
        """Get list of elements in expanded sequence."""
        return ElementList(self._name, self._libmadx, expanded=True)


class ElementList(collections.Sequence):

    """
    Immutable list of beam line elements.

    Each element is a dictionary containing its properties.
    """

    def __init__(self, sequence_name, libmadx, expanded):
        """
        Initialize instance.
        """
        self._sequence_name = sequence_name
        if expanded:
            self._get_element = libmadx.get_expanded_element
            self._get_element_count = libmadx.get_expanded_element_count
            self._get_element_index = libmadx.get_expanded_element_index
            self._get_element_at = libmadx.get_expanded_element_index_by_position
        else:
            self._get_element = libmadx.get_element
            self._get_element_count = libmadx.get_element_count
            self._get_element_index = libmadx.get_element_index
            self._get_element_at = libmadx.get_element_index_by_position

    def __contains__(self, element):
        """
        Check if sequence contains element with specified name.

        Can be invoked with either the element dict or the element name.
        """
        try:
            self.index(element)
            return True
        except ValueError:
            return False

    def __getitem__(self, index):
        """Return element with specified index."""
        if isinstance(index, (dict, basestring)):
            # allow element names to be passed for convenience:
            index = self.index(index)
        # _get_element accepts indices in the range [0, len-1]. The following
        # extends the accepted range to [-len, len+1], just like for lists:
        if index < 0:
            index += len(self)
        return self._get_element(self._sequence_name, index)

    def __len__(self):
        """Get number of elements."""
        return self._get_element_count(self._sequence_name)

    def index(self, element):
        """
        Find index of element with specified name.

        Can be invoked with either the element dict or the element name.

        :raises ValueError:
        """
        if isinstance(element, dict):
            name = element['name']
        else:
            name = element
        return self._get_element_index(self._sequence_name, name)

    def at(self, pos):
        """Find the element at specified S position."""
        return self._get_element_at(self._sequence_name, pos)


class Dict(dict):
    pass


class TableProxy(collections.Mapping):

    """
    Proxy object for lazy-loading table column data.
    """

    def __init__(self, name, libmadx, _check=True):
        """Just store the table name for now."""
        self._name = name
        self._libmadx = libmadx
        if _check and not libmadx.table_exists(name):
            raise ValueError("Invalid table: {!r}".format(name))

    def __getitem__(self, column):
        """Get the column data."""
        try:
            return self._libmadx.get_table_column(self._name, column.lower())
        except ValueError:
            raise KeyError(column)

    def __iter__(self):
        """Iterate over all column names."""
        return iter(self._libmadx.get_table_column_names(self._name))

    def __len__(self):
        """Return number of columns."""
        return len(self._libmadx.get_table_column_names(self._name))

    @property
    def summary(self):
        """Get the table summary."""
        return self._libmadx.get_table_summary(self._name)

    def copy(self, columns=None):
        """
        Return a frozen table with the desired columns.

        :param list columns: column names or ``None`` for all columns.
        :returns: column data
        :rtype: dict
        :raises ValueError: if the table name is invalid
        """
        if columns is None:
            columns = self
        table = Dict((column, self[column]) for column in columns)
        table.summary = self.summary
        return table



# cython: embedsignature=True
"""
Low level cython binding to MAD-X.

CAUTION: Do not import this module directly! Use :class:`Madx` instead.

- Importing this module means loading MAD-X directly into the process space.
  This means that any crash in the (extremely fragile) MAD-X interpreter will
  crash the importing process with it.
- All functions in this module operate on a shared global state.
- this module exposes a very C-ish API that is not convenient to work with.
"""

from os import chdir, getcwd

import ctypes
import numpy as np      # Import the Python-level symbols of numpy

# Import a large-enough integer type to hold pointer, see also:
# http://grokbase.com/t/gg/cython-users/134b21rga8/passing-callback-pointers-to-python-and-back
cdef extern from "pyport.h":
    ctypedef int Py_intptr_t

from cpymad.types import Constraint, Expression
cimport cpymad.clibmadx as clib


# Remember whether start() was called
_madx_started = False


# Python-level binding to libmadx:
__all__ = [
    # MAD-X version
    'get_version_number',
    'get_version_date',

    # control the interpreter
    'is_started',
    'start',
    'finish',
    'input',
    'evaluate',

    # iterate sequences
    'sequence_exists',
    'get_sequence_names',
    'get_sequence_count',
    'get_active_sequence_name',

    # sequence access
    'get_sequence_twiss_table_name',
    'get_sequence_beam',
    'is_sequence_expanded',

    # iterate tables
    'table_exists',
    'get_table_names',
    'get_table_count',

    # table access
    'get_table_summary',
    'get_table_column_names',
    'get_table_column',

    # sequence element list access
    'get_element',
    'get_element_index',
    'get_element_index_by_position',
    'get_element_count',

    # expanded sequence element access
    'get_expanded_element',
    'get_expanded_element_index',
    'get_expanded_element_index_by_position',
    'get_expanded_element_count',

    # these are imported from 'os' for convenience in madx.Madx and should
    # not really be considered part of the public interface:
    'chdir',
    'getcwd',
]


def _get_rightmost_word(sentence):
    """Get the work right of the rightmost space character."""
    return sentence.rsplit(' ', 1)[-1]


def get_version_number():
    """
    Get the version number of loaded MAD-X interpreter.

    :returns: full version number
    :rtype: str
    """
    return _get_rightmost_word(_str(clib.version_name))


def get_version_date():
    """
    Get the release date of loaded MAD-X interpreter.

    :returns: release date in YYYY.MM.DD format
    :rtype: str
    """
    return _get_rightmost_word(_str(clib.version_date))


def is_started():
    """
    Check whether MAD-X has been initialized.

    :returns: whether :func:`start` was called without matching :func:`finish`
    :rtype: bool
    """
    return _madx_started


def start():
    """
    Initialize MAD-X.
    """
    clib.madx_start()
    global _madx_started
    _madx_started = True


def finish():
    """
    Cleanup MAD-X.
    """
    clib.madx_finish()
    global _madx_started
    _madx_started = False


def input(cmd):
    """
    Pass one input command to MAD-X.

    :param str cmd: command to be executed by the MAD-X interpreter
    """
    cdef bytes _cmd = _cstr(cmd)
    clib.stolower_nq(_cmd)
    clib.pro_input(_cmd)


def sequence_exists(sequence_name):
    """
    Check if the sequence exists.

    :param str sequence: sequence name
    :returns: True if the sequence exists
    :rtype: bool
    """
    try:
        _find_sequence(sequence_name)
        return True
    except ValueError:
        return False


def get_sequence_twiss_table_name(sequence_name):
    """
    Get the last calculated twiss table for the given sequence.

    :param str sequence_name: sequence name
    :returns: twiss table name
    :rtype: str
    :raises ValueError: if the sequence name is invalid
    :raises RuntimeError: if the twiss table is invalid
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    if not seq.tw_valid:
        raise RuntimeError("TWISS table invalid.")
    return _str(seq.tw_table.name)


def get_sequence_beam(sequence_name):
    """
    Get the beam associated to the sequence.

    :param str sequence_name: sequence name
    :returns: beam properties as set with the BEAM command (and some more)
    :rtype: dict
    :raises ValueError: if the sequence name is invalid
    :raises RuntimeError: if the sequence has no associated beam
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    if seq.beam is NULL or not seq.beam.beam_def:
        raise RuntimeError("No beam attached to {}".format(sequence_name))
    return _parse_command(seq.beam)


def get_active_sequence_name():
    """
    Get the name of the active sequence.

    :returns: name of active sequence
    :rtype: str
    :raises RuntimeError: if no sequence is activated
    """
    if clib.current_sequ is NULL:
        raise RuntimeError("No active sequence!")
    return _str(clib.current_sequ.name)


def get_sequence_names():
    """
    Get a list of all sequences currently in memory.

    :returns: sequence names
    :rtype: list
    """
    cdef clib.sequence_list* seqs = clib.madextern_get_sequence_list()
    cdef int i
    return [_str(seqs.sequs[i].name)
            for i in xrange(seqs.curr)]


def get_sequence_count():
    """
    Get the number of all sequences currently in memory.

    :returns: number of sequences
    :rtype: int
    """
    return clib.madextern_get_sequence_list().curr


def table_exists(table_name):
    """
    Check if the table exists.

    :param str table: table name
    :returns: True if the table exists
    :rtype: bool
    """
    cdef bytes _table_name = _cstr(table_name)
    return bool(clib.table_exists(_table_name))


def get_table_names():
    """
    Return list of all table names.

    :returns: table names
    :rtype: list
    """
    return [_str(clib.table_register.names.names[i])
            for i in xrange(clib.table_register.names.curr)]


def get_table_count():
    """
    Return number of existing tables.

    :returns: number of tables in memory
    :rtype: int
    """
    return clib.table_register.names.curr


def get_table_summary(table_name):
    """
    Get table summary.

    :param str table_name: table name
    :returns: mapping of {column: value}
    :rtype: dict
    """
    cdef bytes _table_name = _cstr(table_name)
    cdef clib.char_p_array* header = clib.table_get_header(_table_name)
    cdef int i
    if header is NULL:
        raise ValueError("No summary for table: {!r}".format(table_name))
    return dict([_split_header_line(header.p[i])
                 for i in xrange(header.curr)])


def get_table_column_names(table_name):
    """
    Get a list of all column names in the table.

    :param str table_name: table name
    :returns: column names
    :rtype: list
    :raises ValueError: if the table name is invalid
    """
    cdef bytes _table_name = _cstr(table_name)
    cdef int index = clib.name_list_pos(_table_name, clib.table_register.names)
    if index == -1:
        raise ValueError("Invalid table: {!r}".format(table_name))
    # NOTE: we can't enforce lower-case on the column names here, since this
    # breaks their usage with get_table_column():
    return _name_list(clib.table_register.tables[index].columns)


def get_table_column_count(table_name):
    """
    Get a number of columns in the table.

    :param str table_name: table name
    :returns: number of columns
    :rtype: int
    :raises ValueError: if the table name is invalid
    """
    cdef bytes _table_name = _cstr(table_name)
    cdef int index = clib.name_list_pos(_table_name, clib.table_register.names)
    if index == -1:
        raise ValueError("Invalid table: {!r}".format(table_name))
    return clib.table_register.tables[index].columns.curr


def get_table_column(table_name, column_name):
    """
    Get data from the specified table.

    :param str table_name: table name
    :param str column_name: column name
    :returns: the data in the requested column
    :rtype: numpy.array
    :raises ValueError: if the column cannot be found in the table
    :raises RuntimeError: if the column has unknown type

    CAUTION: Numeric data is wrapped in numpy arrays but not copied. Make
    sure to copy all data before invoking any further MAD-X commands! This
    is done automatically for you if using libmadx in a remote service
    (pickle serialization effectively copies the data).
    """
    cdef char** char_tmp
    cdef bytes _tab_name = _cstr(table_name)
    cdef bytes _col_name = _cstr(column_name)
    cdef clib.column_info info
    # The following snippet imitates the table_get_column function from the
    # C API of the MAD-X library. The replacement is needed for proper
    # handling of integer columns for as long as we support versions of
    # MAD-X before r4777. When this is not the case anymore, you should be
    # able to safely revert the commit that introduced this snippet:
    cdef int tab_i = clib.name_list_pos(_tab_name, clib.table_register.names)
    if tab_i == -1:
        raise ValueError("Invalid table: {!r}".format(table_name))
    cdef clib.table* _table = clib.table_register.tables[tab_i]
    cdef int col_i = clib.name_list_pos(_col_name, _table.columns)
    if col_i == -1:
        raise ValueError("Invalid column: {!r}".format(column_name))
    info.length = _table.curr
    cdef int inform = _table.columns.inform[col_i]
    if inform == clib.PARAM_TYPE_INTEGER:
        # YES, integers are internally stored as doubles in MAD-X:
        info.data = _table.d_cols[col_i]
        info.datatype = b'i'
    elif inform == clib.PARAM_TYPE_DOUBLE:
        info.data = _table.d_cols[col_i]
        info.datatype = b'd'
    elif inform == clib.PARAM_TYPE_STRING:
        info.data = _table.s_cols[col_i]
        info.datatype = b'S'
    else:
        raise RuntimeError("Unknown column format: {!r}".format(inform))
    # Although, we are using a custom replacement for table_get_column
    # above, we still try to be fully compatible with the real function from
    # MAD-X below, so back-migration will be easier, when the time comes.
    # This is why the error and data type handling below is left untouched:
    dtype = <bytes> info.datatype
    size = <int> info.length
    addr = <Py_intptr_t> info.data
    # double:
    if dtype == b'i' or dtype == b'd':
        # YES, integers are internally stored as doubles in MAD-X:
        array_type = ctypes.c_double * size
        array_data = array_type.from_address(addr)
        return np.ctypeslib.as_array(array_data)
    # string:
    elif dtype == b'S':
        char_tmp = <char**> info.data
        return np.array([char_tmp[i] for i in xrange(info.length)])
    # invalid:
    elif dtype == b'V':
        raise ValueError("Column {!r} is not in table {!r}."
                         .format(column_name, table_name))
    # unknown:
    else:
        raise RuntimeError("Unknown datatype {!r} in column {!r}."
                           .format(_str(dtype), column_name))


def get_element(sequence_name, element_index):
    """
    Return requested element in the original sequence.

    :param str sequence_name: sequence name
    :param int element_index: element index
    :returns: the element with the specified index
    :rtype: dict
    :raises ValueError: if the sequence is invalid
    :raises IndexError: if the index is out of range
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    if element_index < 0 or element_index >= seq.nodes.curr:
        raise IndexError("Index out of range: {0} (element count is {1})"
                         .format(element_index, seq.nodes.curr))
    return _get_node(seq.nodes.nodes[element_index], seq.ref_flag)


def get_element_index(sequence_name, element_name):
    """
    Return index of element with specified name in the original sequence.

    :param str sequence_name: sequence name
    :param str element_name: element index
    :returns: the index of the specified element
    :rtype: int
    :raises ValueError: if the sequence or element name is invalid
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef bytes _element_name = _cstr(element_name)
    cdef int index = clib.name_list_pos(_element_name, seq.nodes.list)
    if index == -1:
        raise ValueError("Element name not found: {0}".format(element_name))
    return index


def get_element_index_by_position(sequence_name, position):
    """
    Return index of element at specified position in the original sequence.

    :param str sequence_name: sequence name
    :param double position: position (S coordinate)
    :returns: the index of an element at that position, -1 if not found
    :rtype: int
    :raises ValueError: if the sequence or element name is invalid
    """
    # This is implemented here just for performance, but still uses a linear
    # algorithm. If you want more, you need to copy the elements over and use
    # a suitable ordered data type.
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef double _position = position
    cdef clib.node* elem
    cdef double at
    for i in xrange(seq.nodes.curr):
        elem = seq.nodes.nodes[i]
        at = _get_node_entry_pos(elem, seq.ref_flag)
        if _position >= at and _position <= at+elem.length:
            return i
    raise ValueError("No element found at position: {0}".format(position))


def get_element_count(sequence_name):
    """
    Return number of elements in the original sequence.

    :param str sequence_name: sequence name
    :returns: number of elements in the original sequence
    :rtype: int
    :raises ValueError: if the sequence is invalid.
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    return seq.nodes.curr


def get_expanded_element(sequence_name, element_index):
    """
    Return requested element in the expanded sequence.

    :param str sequence_name: sequence name
    :param int element_index: element index
    :returns: the element with the specified index
    :rtype: dict
    :raises ValueError: if the sequence is invalid
    :raises IndexError: if the index is out of range
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    if element_index < 0 or element_index >= seq.n_nodes:
        raise IndexError("Index out of range: {0} (element count is {1})"
                         .format(element_index, seq.n_nodes))
    return _get_node(seq.all_nodes[element_index], seq.ref_flag)


def get_expanded_element_index(sequence_name, element_name):
    """
    Return index of element with specified name in the expanded sequence.

    NOTE: this is the brute-force linear-time algorithm and therefore not
    recommended for frequent execution.

    :param str sequence_name: sequence name
    :param str element_name: element index
    :returns: the index of the specified element, -1 if not found
    :rtype: int
    :raises ValueError: if the sequence is invalid
    """
    # Unfortunately, there is no name_list for the expanded list. Therefore,
    # Therefore, we can only provide a linear-time lookup.
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef bytes _element_name = _cstr(element_name)
    for i in xrange(seq.n_nodes):
        if seq.all_nodes[i].name == _element_name:
            return i
    raise ValueError("Element name not found: {0}".format(element_name))


def get_expanded_element_index_by_position(sequence_name, position):
    """
    Return index of element at specified position in the expanded sequence.

    :param str sequence_name: sequence name
    :param double position: position (S coordinate)
    :returns: the index of an element at that position
    :rtype: int
    :raises ValueError: if the sequence or element name is invalid
    """
    # This is implemented here just for performance, but still uses a linear
    # algorithm. If you want more, you need to copy the elements over and use
    # a suitable ordered data type.
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    cdef double _position = position
    cdef clib.node* elem
    cdef double at
    for i in xrange(seq.n_nodes):
        elem = seq.all_nodes[i]
        at = _get_node_entry_pos(elem, seq.ref_flag)
        if _position >= at and _position <= at+elem.length:
            return i
    raise ValueError("No element found at position: {0}".format(position))


def get_expanded_element_count(sequence_name):
    """
    Return number of elements in the expanded sequence.

    :param str sequence_name: sequence name
    :returns: number of elements in the expanded sequence
    :rtype: int
    :raises ValueError: if the sequence is invalid.
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    return seq.n_nodes


def is_sequence_expanded(sequence_name):
    """
    Check whether a sequence has already been expanded.

    :param str sequence_name: sequence name
    :returns: expanded state of the sequence
    :rtype: bool
    :raises ValueError: if the sequence is invalid
    """
    cdef clib.sequence* seq = _find_sequence(sequence_name)
    return seq.n_nodes > 0


def evaluate(expression):
    """
    Evaluates an expression and returns the result as double.

    :param str expression: symbolic expression to evaluate
    :returns: numeric value of the expression
    :rtype: float
    """
    # TODO: This function uses global variables as temporaries - which is in
    # general an *extremely* bad design choice. Even though MAD-X uses global
    # variables internally anyway, this is no excuse for cpymad to copy that
    # behaviour.
    try:
        # handle instance of type Expression:
        expression = expression.expr
    except AttributeError:
        pass
    # TODO: not sure about the flags (the magic constants 0, 2)
    cdef bytes _expr = _cstr(expression.lower())
    clib.pre_split(_expr, clib.c_dum, 0)
    clib.mysplit(clib.c_dum.c, clib.tmp_p_array)
    expr = clib.make_expression(clib.tmp_p_array.curr, clib.tmp_p_array.p)
    value = clib.expression_value(expr, 2)
    clib.delete_expression(expr)
    return value


# Helper functions:

# The following functions are `cdef functions`, i.e. they can only be
# called from Cython code. It is necessary to use `cdef functions` whenever
# we want to pass parameters or return values with a pure C type.

_expr_types = [bool, int, float]

cdef _expr(clib.expression* expr,
           double value,
           int typeid=clib.PARAM_TYPE_DOUBLE):
    """Return a parameter value with an appropriate type."""
    _type = _expr_types[typeid]
    _value = _type(value)
    if expr is NULL or expr.string is NULL:
        return _value
    _expr = _str(expr.string).strip()
    if not _expr:
        return _value
    try:
        return _type(float(_expr))
    except ValueError:
        pass
    return Expression(_expr, _type(evaluate(_expr)), _type)


cdef _get_param_value(clib.command_parameter* par):

    """
    Get the value of a command parameter.

    :returns: value of the parameter
    :rtype: depends on the parameter type
    :raises ValueError: if the parameter type is invalid
    """

    if par.type in (clib.PARAM_TYPE_LOGICAL,
                    clib.PARAM_TYPE_INTEGER,
                    clib.PARAM_TYPE_DOUBLE):
        return _expr(par.expr, par.double_value, par.type)

    if par.type == clib.PARAM_TYPE_STRING:
        return _str(par.string)

    if par.type == clib.PARAM_TYPE_CONSTRAINT:
        if par.c_type == clib.CONSTR_TYPE_MIN:
            return Constraint(min=_expr(par.min_expr, par.c_min))
        if par.c_type == clib.CONSTR_TYPE_MAX:
            return Constraint(max=_expr(par.max_expr, par.c_max))
        if par.c_type == clib.CONSTR_TYPE_BOTH:
            return Constraint(min=_expr(par.min_expr, par.c_min),
                              max=_expr(par.max_expr, par.c_max))
        if par.c_type == clib.CONSTR_TYPE_VALUE:
            return Constraint(val=_expr(par.expr, par.double_value))

    if par.type in (clib.PARAM_TYPE_INTEGER_ARRAY, clib.PARAM_TYPE_DOUBLE_ARRAY):
        return [
            _expr(NULL if par.expr_list is NULL else par.expr_list.list[i],
                  par.double_array.a[i],
                  par.type - clib.PARAM_TYPE_LOGICAL_ARRAY)
            for i in xrange(par.double_array.curr)
        ]

    if par.type == clib.PARAM_TYPE_STRING_ARRAY:
        return [_str(par.m_string.p[i])
                for i in xrange(par.m_string.curr)]

    raise ValueError("Unknown parameter type: {}".format(par.type))


cdef _parse_command(clib.command* cmd):
    """
    Get the values of all parameters of a command.

    :returns: the command parameters
    :rtype: dict
    """
    # generator expressions are not yet supported in cdef functions, so
    # let's do it the hard way:
    res = {}
    cdef int i
    for i in xrange(cmd.par.curr):
        # enforce lower-case keys:
        name = _str(cmd.par.parameters[i].name).lower()
        res[name] = _get_param_value(cmd.par.parameters[i])
    return res


# The 'except NULL' clause is needed to forward exceptions from cdef
# functions with C return values, see:
# http://docs.cython.org/src/userguide/language_basics.html#error-return-values
cdef clib.sequence* _find_sequence(sequence_name) except NULL:
    """
    Get pointer to the C sequence struct of the specified sequence or NULL.

    :param str sequence_name: sequence name
    :raises ValueError: if the sequence can not be found
    """
    cdef bytes _sequence_name = _cstr(sequence_name)
    cdef clib.sequence_list* seqs = clib.madextern_get_sequence_list()
    cdef int index = clib.name_list_pos(_sequence_name, seqs.list)
    if index == -1:
        raise ValueError("Invalid sequence: {}".format(sequence_name))
    return seqs.sequs[index]


cdef _split_header_line(header_line):
    """Parse a table header value."""
    _, key, kind, value = _str(header_line).split(None, 3)
    key = key.lower()
    if kind == "%le":
        return key, float(value)    # convert to number
    elif kind.endswith('s'):
        return key, value[1:-1]     # strip quotes from string
    else:
        return key, value           #


cdef _name_list(clib.name_list* names):
    """Return a python list of names for the name_list."""
    cdef int i
    return [names.names[i] for i in xrange(names.curr)]


cdef _str(char* s):
    """Decode C string to python string."""
    if s is NULL:
        # Returning an empty string will make the type of the parameter
        # transparent to the inspecting code:
        return ""
    return s.decode('utf-8')


cdef bytes _cstr(s):
    """Encode python string to C string."""
    if s is None:
        return b""
    return <bytes> s.encode('utf-8')


cdef _get_node(clib.node* node, int ref_flag):
    """Return dictionary with node + element attributes."""
    if node.p_elem is NULL:
        # Maybe this is a valid case, but better detect it with boom!
        raise RuntimeError("Empty node or subsequence! Please report this incident!")
    data = _get_element(node.p_elem)
    data.update({'name': _str(node.name),
                 'type': _str(node.base_name),
                 'at': _get_node_entry_pos(node, ref_flag)})
    return data


cdef double _get_node_entry_pos(clib.node* node, int ref_flag):
    """Normalize 'at' value to node entry."""
    if ref_flag == clib.REF_CENTER:
        return node.at_value - node.length / 2
    elif ref_flag == clib.REF_EXIT:
        return node.at_value - node.length
    else:
        return node.at_value


cdef _get_element(clib.element* elem):
    """Return dictionary with element attributes."""
    return _parse_command(elem.def_)



# encoding: utf-8
"""
Contains base class for resource providers.
"""

import os
from contextlib import contextmanager
from shutil import copyfileobj
from tempfile import NamedTemporaryFile


__all__ = [
    'ResourceProvider',
]


class ResourceProvider(object):

    """
    Abstract base class for resource providers.

    Resources are read-only (at the moment) data objects such as
    model-data. Resource providers have a common interface to abstract the
    underlying API to access those resources.
    """

    def open(self, name='', encoding=None):
        """
        Open the specified resource.

        :param string name: Name of the resource, optional.
        :param string encoding: Either None or encoding to use (optional).

        Returns a file-like object. When finished using ``close()`` must be
        called on the returned object. If ``encoding`` is ``None`` the
        stream is opened in binary mode.
        """
        raise NotImplementedError("ResourceProvider.open")

    def listdir(self, name=''):
        """
        List directory contents.

        :param string name: Name of the resource, optional.

        This works similar to os.listdir().
        restrict can be used to filter for certain file types.
        """
        raise NotImplementedError("ResourceProvider.listdir")

    def get(self, name):
        """
        Get a provider object relative to the specified subdirectory.

        :param string name: Name of the resource, optional.

        Returns an instance of the ResourceProvider that opens, lists and
        gets objects relative to the specified subdirectory.
        """
        raise NotImplementedError("ResourceProvider.get")

    def provider(self):
        """
        Get the parent provider object.
        """
        raise NotImplementedError("ResourceProvider.provider")

    # mixins:
    def listdir_filter(self, name='', ext=''):
        """
        List resources that match restriction.

        :param string name: Name of the resource, optional.
        :param string ext: Filename extension to be filtered (including dot).

        NOTE: To stay upward compatible ext should only be passed as keyword
        argument.
        """
        for res_name in self.listdir(name):
            if res_name.lower().endswith(ext):
                yield res_name

    def load(self, name='', encoding=None):
        """
        Load the specified resource into memory and return all the data.

        :param string name: Name of the resource, optional.
        :param string encoding: Either None or encoding to use (optional).

        If ``encoding`` is ``None`` the returned data is binary.
        This is a convenience mixin.
        """
        with self.open(name, encoding) as f:
            return f.read()

    def yaml(self, name='', encoding='utf-8', **kwargs):
        """
        Load the specified YAML/JSON resource.

        :param string name: Name of the resource, optional.
        :param string encoding: Encoding to use.

        kwargs can be used to pass additional arguments to the YAML parser.
        This is a convenience mixin.

        Note that a YAML parser is used but since JSON is a subset of YAML
        this function can also be used to load JSON resources. The input is
        not checked to be valid JSON!
        """
        from yaml import safe_load
        with self.open(name, encoding=encoding) as f:
            return safe_load(f, **kwargs)

    # backward compatibility alias
    json = yaml

    @contextmanager
    def filename(self, name=''):
        """
        Yield the path of a file containing the resource.

        Use this as a context manager to make sure temporary files are
        deleted when done using. Example:

        .. code-block:: python

            res = PackageResource('foo_module_123') # may .egg or filesystem
            with res.filename('foo.txt') as filename:
                with open(filename) as file:
                    content = file.read()
            # temporarily extracted files are deleted at this point
        """
        try:
            tempfile = tempfile.NamedTemporaryFile(mode='wb', delete=False)
            with tempfile as dest:
                with self.open(name) as src:
                    copyfileobj(src, dest)
            yield tempfile.name
        finally:
            os.remove(tempfile.name)



# encoding: utf-8
from __future__ import unicode_literals

__title__ = 'cpymad'
__version__ = '0.10.2'

__summary__ = 'Cython binding to MAD-X'
__uri__ = 'https://github.com/hibtc/cpymad'

__author__ = 'PyMAD developers'
__author_email__ = 'pymad@cern.ch'
__maintainer__ = 'Thomas Gläßle'
__maintainer_email__ = 't_glaessle@gmx.de'

__license__ = 'CC0, Apache, Non-Free'
__copyright__ = 'See file COPYING.rst'

# Trove classifiers: https://pypi.python.org/pypi?:action=list_classifiers
__classifiers__ = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Science/Research',
    'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
    'License :: OSI Approved :: Apache Software License',
    'License :: Other/Proprietary License',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.4',
    'Topic :: Scientific/Engineering :: Physics',
]



# standard library
import unittest

# utilities
import _compat

# tested objects
from cpymad import util


class TestUtil(unittest.TestCase, _compat.TestCase):

    """Tests for the objects in :mod:`cpymad.util`."""

    def test_is_identifier(self):
        self.assertTrue(util.is_identifier('hdr23_oei'))
        self.assertTrue(util.is_identifier('_hdr23_oei'))
        self.assertFalse(util.is_identifier('2_hdr23_oei'))
        self.assertFalse(util.is_identifier('hdr oei'))
        self.assertFalse(util.is_identifier('hdr@oei'))

    def test_strip_element_suffix(self):
        self.assertEqual(util.strip_element_suffix('foo.23'), 'foo.23')
        self.assertEqual(util.strip_element_suffix('foo.23:43'), 'foo.23')
        # TODO: this should raise:
        self.assertEqual(util.strip_element_suffix('foo:23o'), 'foo:23o')

    def test_add_element_suffix(self):
        self.assertEqual(util.add_element_suffix('foo.23'), 'foo.23:1')
        self.assertEqual(util.add_element_suffix('foo.23:43'), 'foo.23:43')
        # TODO: this should raise:
        self.assertEqual(util.add_element_suffix('foo:23o'), 'foo:23o:1')

    def test_normalize_range_name(self):
        self.assertEqual(util.normalize_range_name('dr:1'), 'dr')
        self.assertEqual(util.normalize_range_name('lebt$end'), '#e')
        self.assertEqual(util.normalize_range_name('dr'), 'dr')

    # TODO: test other functions


if __name__ == '__main__':
    unittest.main()



# encoding: utf-8
"""
Tests for the model.Model runtime hierarchy.
"""

# tested classes
from cpymad.madx import CommandLog
from cpymad.model import Model
from cpymad.resource.file import FileResource

# utilities
import _compat

# standard library
import os
import sys
import unittest

__all__ = [
    'TestModel',
]


class TestModel(unittest.TestCase, _compat.TestCase):

    """
    Tests for the Model class.
    """

    # test configuration:

    path = os.path.join(os.path.dirname(__file__), 'data', 'lebt.cpymad.yml')

    # helper methods for tests:

    def load_model(self, path):
        """Load model with given name from specified path."""
        command_log = CommandLog(sys.stdout, 'X:> ')
        model = Model.load(self.path, command_log=command_log)
        model.madx.command.option(twiss_print=False)
        return model

    def setUp(self):
        self.model = self.load_model(self.path)

    def tearDown(self):
        del self.model

    # tests for Model API

    def test_compatibility_check(self):
        data = {
            'beams': {},
            'optics': {},
            'sequences': {},
        }
        with self.assertRaises(ValueError):
            Model(data=data, repo=None, madx=None)
        with self.assertRaises(ValueError):
            Model(data=dict(data, api_version=-1), repo=None, madx=None)
        with self.assertRaises(ValueError):
            Model(data=dict(data, api_version=1), repo=None, madx=None)
        Model(data=dict(data, api_version=0), repo=None, madx=None)

    def test_Model_API(self):
        """Check that the public Model attributes/methods behave reasonably."""
        model = self.model
        madx = model.madx
        # name
        self.assertEqual(model.name, 'lebt')
        # data
        repository = FileResource(self.path)
        self.assertEqual(model.data, repository.yaml())
        # beams
        self.assertItemsEqual(model.beams.keys(), ['carbon', 'other'])
        # optics
        self.assertItemsEqual(model.optics.keys(), ['thick'])
        # sequences
        self.assertItemsEqual(model.sequences.keys(), ['s1', 's2'])
        self.assertTrue('s1' in madx.sequences)
        self.assertTrue('s2' in madx.sequences)
        # default_optic
        self.assertIs(model.default_optic, model.optics['thick'])
        # default_sequence
        self.assertIs(model.default_sequence, model.sequences['s1'])

    # tests for Optic API

    def test_Optic_load(self):
        """Check that the Optic init-files are executed upon init()."""
        evaluate = self.model.madx.evaluate
        self.assertAlmostEqual(evaluate('NOT_ZERO'), 0.0)
        self.model.optics['thick'].init()
        self.assertAlmostEqual(evaluate('NOT_ZERO'), 1.0)

    # tests for Sequence API

    def test_Sequence_API(self):
        """Check that the general Sequence API behaves reasonable."""
        sequence = self.model.sequences['s1']
        # name
        self.assertEqual(sequence.name, 's1')
        # ranges
        self.assertItemsEqual(sequence.ranges.keys(), ['ALL'])
        # beam
        self.assertIs(sequence.beam, self.model.beams['carbon'])
        # default_range
        self.assertIs(sequence.default_range, sequence.ranges['ALL'])

    # def test_Sequence_twiss(self):        # see test_Optic_twiss for now
    # def test_Sequence_match(self):        # see test_Optic_match for now

    def test_Sequence_survey(self):
        """Execute survey() and check that it returns usable values."""
        seq = self.model.default_sequence
        survey = seq.survey()
        # access some data to make sure the table was generated:
        s = survey['s']
        x = survey['x']
        y = survey['y']
        z = survey['z']

    # tests for Range API

    def test_Range_API(self):
        """Check that the general Range API behaves reasonable."""
        range = self.model.default_sequence.default_range
        # name
        self.assertEqual(range.name, 'ALL')
        # bounds
        self.assertEqual(range.bounds, ('#s', '#e'))
        # initial_conditions
        self.assertItemsEqual(range.initial_conditions.keys(), ['default'])
        # default_initial_conditions
        self.assertIs(range.default_initial_conditions,
                      range.initial_conditions['default'])

    def test_Range_twiss(self):
        """Execute twiss() and check that it returns usable values."""
        range = self.model.default_sequence.default_range
        twiss = range.twiss()
        # access some data to make sure the table was generated:
        betx = twiss['betx']
        bety = twiss['bety']
        alfx = twiss['alfx']
        alfy = twiss['alfy']

    def test_Range_match(self):
        """Execute match() and check that it returns usable values."""
        range = self.model.sequences['s1'].default_range
        knobs = range.match(
            constraints=[dict(range='sb', betx=0.3)],
            vary=['QP_K1'],
        )
        knobs['QP_K1']


if __name__ == '__main__':
    unittest.main()



"""
Simple RPC module for libmadx.

This module is needed to execute several instances of the `libmadx` module
in remote processes and communicate with them via remote procedure calls
(RPC). Use the :meth:`LibMadxClient.spawn` to create a new instance.

The remote backend is needed due to the fact that cpymad.libmadx is a low
level binding to the MAD-X library which in turn uses global variables.
This means that the cpymad.libmadx module has to be loaded within remote
processes in order to deal with several isolated instances of MAD-X in
parallel.

Furthermore, this can be used as a security enhancement: if dealing with
unverified input, we can't be sure that a faulty MAD-X function
implementation will give access to a secure resource. This can be executing
all library calls within a subprocess that does not inherit any handles.

More importantly: the MAD-X program crashes on the tinyest error. Boxing it
in a subprocess will prevent the main process from crashing as well.

CAUTION:

The service communicates with the remote end via pickling, i.e. both ends
can execute arbitrary code on the other side. This means that the remote
process can not be used to safely execute unsafe python code.
"""

from __future__ import absolute_import

__all__ = [
    'LibMadxClient',
    'RemoteProcessCrashed',
]

import traceback
import os
import subprocess
import sys

try:
    # python2's cPickle is an accelerated (C extension) version of pickle:
    import cPickle as pickle
except ImportError:
    # python3's pickle automatically uses the accelerated version and falls
    # back to the python version, see:
    # http://docs.python.org/3.3/whatsnew/3.0.html?highlight=cpickle
    import pickle


_win = sys.platform == 'win32'


def _nop(x):
    """NO-OP: do nothing, just return x."""
    return x


class RemoteProcessClosed(RuntimeError):
    """The MAD-X remote process has already been closed."""
    pass


class RemoteProcessCrashed(RuntimeError):
    """The MAD-X remote process has crashed."""
    pass


if _win:
    import msvcrt


    try:                    # python2
        import _subprocess as _winapi
        # _subprocess.DuplicateHandle and _subprocess.CreatePipe return a
        # handle type with .Close() and .Detach() methods:
        Handle = _nop
    except ImportError:     # python3
        import _winapi
        # _winapi.DuplicateHandle and _winapi.CreatePipe return plain
        # integers which need to be wrapped in subprocess.Handle to make
        # them closable:
        Handle = subprocess.Handle


    def _make_inheritable(handle):
        """Return inheritable Handle, close the original."""
        # new handles are created uninheritable by default, but they can be
        # made inheritable on duplication:
        current_process = _winapi.GetCurrentProcess()
        dup = Handle(_winapi.DuplicateHandle(
            current_process,
            handle,
            current_process,
            0,
            True,
            _winapi.DUPLICATE_SAME_ACCESS))
        handle.Close()
        return dup


    def _pipe():
        """Create a unidirectional pipe."""
        # use _winapi.CreatePipe on windows, just like subprocess.Popen
        # does when requesting PIPE streams. This is the easiest and most
        # reliable method I have tested so far:
        recv, send = _winapi.CreatePipe(None, 0)
        return Handle(recv), Handle(send)


    def _open(handle):
        """
        Open a file descriptor for the specified HANDLE (int -> int).

        Closing the file descriptor will also close the handle.
        """
        return msvcrt.open_osfhandle(handle, 0)


    def _close(handle):
        """Close the :class:`Handle` object."""
        handle.Close()


    def _detach(handle):
        """Return HANDLE after detaching it from a :class:`Handle` object."""
        return handle.Detach()


# POSIX
else:
    try:
        from os import set_inheritable
    except ImportError:         # python2
        # on POSIX/python2 file descriptors are inheritable by default:
        _make_inheritable = _nop
    else:                       # python3
        def _make_inheritable(fd):
            """Return inheritable file descriptor, close the original."""
            dup = os.dup(fd)
            os.close(fd)
            set_inheritable(dup, True)
            return dup


    _pipe = os.pipe


    # handles are just file descriptors on POSIX:
    _open = _nop
    _close = os.close
    _detach = _nop


def _close_all_but(keep):
    """Close all but the given file descriptors."""
    # first, let the garbage collector run, it may find some unreachable
    # file objects (on posix forked processes) and close them:
    import gc
    gc.collect()
    # close all ranges in between the file descriptors to be kept:
    keep = sorted(set([-1] + keep + [subprocess.MAXFD]))
    for s, e in zip(keep[:-1], keep[1:]):
        if s+1 < e:
            os.closerange(s+1, e)


class Connection(object):

    """
    Pipe-like IPC connection using file objects.

    For most purposes this should behave like the connection objects
    returned by :func:`multiprocessing.Pipe`.

    This class combines two orthogonal functionalities. In general this is
    bad practice, meaning the class should be refactored into two classes,
    but for our specific purpose this will do it.

    - build a bidirectional stream from two unidirectional streams
    - build a serialized connection from pure data streams (pickle)
    """

    def __init__(self, recv, send):
        """Create duplex connection from two unidirectional streams."""
        self._recv = recv
        self._send = send

    def recv(self):
        """Receive a pickled message from the remote end."""
        return pickle.load(self._recv)

    def send(self, data):
        """Send a pickled message to the remote end."""
        # '-1' instructs pickle to use the latest protocol version. This
        # improves performance by a factor ~50-100 in my tests:
        return pickle.dump(data, self._send, -1)

    def close(self):
        """Close the connection."""
        self._recv.close()
        self._send.close()

    @property
    def closed(self):
        """Check if the connection is fully closed."""
        return self._recv.closed and self._send.closed

    @classmethod
    def from_fd(cls, recv_fd, send_fd):
        """Create a connection from two file descriptors."""
        return cls(os.fdopen(recv_fd, 'rb', 0),
                   os.fdopen(send_fd, 'wb', 0))


class Client(object):

    """
    Base class for a very lightweight synchronous RPC client.

    Uses a connection that shares the interface with :class:`Connection` to
    do synchronous RPC. Synchronous IO means that currently callbacks /
    events are impossible.
    """

    def __init__(self, conn):
        """Initialize the client with a :class:`Connection` like object."""
        self._conn = conn

    def __del__(self):
        """Close the client and the associated connection with it."""
        try:
            self.close()
        except (RemoteProcessCrashed, RemoteProcessClosed):
            # catch ugly follow-up warnings after a MAD-X process has crashed
            pass

    @classmethod
    def spawn_subprocess(cls, **Popen_args):
        """
        Create client for a backend service in a subprocess.

        You can use the keyword arguments to pass further arguments to
        Popen, which is useful for example, if you want to redirect STDIO
        streams.
        """
        # create two unidirectional pipes for communication with the
        # subprocess.  the remote end needs to be inheritable, the local
        # end is not inheritable by default (or will be closed by
        # _close_all_but on POSIX/py2):
        local_recv, _remote_send = _pipe()
        _remote_recv, local_send = _pipe()
        remote_recv = _make_inheritable(_remote_recv)
        remote_send = _make_inheritable(_remote_send)
        args = [sys.executable, '-u', '-m', __name__]
        args += [str(int(remote_recv)),
                 str(int(remote_send))]
        proc = subprocess.Popen(args, close_fds=False, **Popen_args)
        # close handles that are not used in this process:
        _close(remote_recv)
        _close(remote_send)
        conn = Connection.from_fd(_open(_detach(local_recv)),
                                  _open(_detach(local_send)))
        return cls(conn), proc

    def close(self):
        """Close the connection gracefully, stop the remote service."""
        try:
            self._conn.send(('close', ()))
        except ValueError:      # already closed
            pass
        self._conn.close()

    @property
    def closed(self):
        """Check if connection is closed."""
        return self._conn.closed

    def _request(self, kind, *args):
        """Communicate with the remote service synchronously."""
        try:
            self._conn.send((kind, args))
        except ValueError:
            if self.closed:
                raise RemoteProcessClosed()
            raise
        except IOError:
            raise RemoteProcessCrashed()
        try:
            response = self._conn.recv()
        except EOFError:
            raise RemoteProcessCrashed()
        return self._dispatch(response)

    def _dispatch(self, response):
        """Dispatch an answer from the remote service."""
        kind, args = response
        handler = getattr(self, '_dispatch_%s' % (kind,))
        return handler(*args)

    def _dispatch_exception(self, exc_info):
        """Dispatch an exception."""
        raise exc_info

    def _dispatch_data(self, data):
        """Dispatch returned data."""
        return data


class Service(object):

    """
    Base class for a very lightweight synchronous RPC service.

    Counterpart to :class:`Client`.
    """

    def __init__(self, conn):
        """Initialize the service with a :class:`Connection` like object."""
        self._conn = conn

    @classmethod
    def stdio_main(cls, args):
        """Do the full job of preparing and running an RPC service."""
        passed_handles = [int(arg) for arg in args]
        _close_all_but([sys.stdin.fileno(),
                        sys.stdout.fileno(),
                        sys.stderr.fileno()] + passed_handles)
        hrecv, hsend = passed_handles[:2]
        conn = Connection.from_fd(_open(hrecv),
                                  _open(hsend))
        cls(conn).run()

    def run(self):
        """
        Run the service until terminated by either the client or user.

        The service is terminated on user interrupts (Ctrl-C), which might
        or might not be desired.
        """
        import logging
        logging.basicConfig(logLevel=logging.INFO)
        logger = logging.getLogger(__name__)
        try:
            while self._communicate():
                pass
        except KeyboardInterrupt:
            logger.info('User interrupt!')
        finally:
            self._conn.close()

    def _communicate(self):
        """
        Receive and serve one RPC request.

        :returns: ``True`` if the service should continue running.
        """
        try:
            request = self._conn.recv()
        except EOFError:
            return False
        else:
            return self._dispatch(request)

    def _dispatch(self, request):
        """
        Dispatch one RPC request.

        :returns: ``True`` if the service should continue running.
        """
        kind, args = request
        handler = getattr(self, '_dispatch_%s' % (kind,))
        try:
            response = handler(*args)
        except:
            self._reply_exception(sys.exc_info())
        else:
            try:
                self._reply_data(response)
            except ValueError:
                if self._conn.closed:
                    return False
                raise
        return True

    def _dispatch_close(self):
        """Close the connection gracefully as initiated by the client."""
        self._conn.close()

    def _reply_data(self, data):
        """Return data to the client."""
        self._conn.send(('data', (data,)))

    def _reply_exception(self, exc_info):
        """Return an exception state to the client."""
        message = exc_info[0](
            "\n" + "".join(traceback.format_exception(*exc_info))),
        self._conn.send(('exception', message))


class LibMadxClient(Client):

    """
    Specialized client for boxing :mod:`cpymad.libmadx` function calls.

    Boxing these MAD-X function calls is necessary due the global nature of
    all state within the MAD-X library.
    """

    def close(self):
        """Finalize libmadx if it is running."""
        try:
            if self.libmadx.is_started():
                self.libmadx.finish()
        except (RemoteProcessClosed, RemoteProcessCrashed):
            pass
        super(LibMadxClient, self).close()

    @property
    def libmadx(self):
        return self.modules['cpymad.libmadx']

    @property
    class modules(object):

        """Provides access to all modules in the remote process."""

        def __init__(self, client):
            self.__client = client

        def __getitem__(self, key):
            """Get a RemoteModule object by module name."""
            return RemoteModule(self.__client, key)


class RemoteModule(object):

    """Wrapper for :mod:`cpymad.libmadx` in a remote process."""

    def __init__(self, client, module):
        """Store the client connection."""
        self.__client = client
        self.__module = module

    def __getattr__(self, funcname):
        """Resolve all attribute accesses as remote function calls."""
        def DeferredMethod(*args, **kwargs):
            return self.__client._request('function_call', self.__module,
                                          funcname, args, kwargs)
        return DeferredMethod


class LibMadxService(Service):

    """
    Specialized service to dispatch :mod:`cpymad.libmadx` function calls.

    Counterpart for :class:`LibMadxClient`.
    """

    def _dispatch_function_call(self, modname, funcname, args, kwargs):
        """Execute any static function call in the remote process."""
        # As soon as we drop support for python2.6, we should replace this
        # with importlib.import_module:
        module = __import__(modname, None, None, '*')
        function = getattr(module, funcname)
        return function(*args, **kwargs)


if __name__ == '__main__':
    LibMadxService.stdio_main(sys.argv[1:])



"""
Python type analogues for MAD-X data structures.
"""

from collections import namedtuple

__all__ = [
    'Range',
    'Constraint',
    'Expression',
]


Range = namedtuple('Range', ['first', 'last'])


class Constraint(object):

    """Represents a MAD-X constraint, which has either min/max/both/value."""

    def __init__(self, val=None, min=None, max=None):
        """Just store the values"""
        self.val = val
        self.min = min
        self.max = max


class Expression(object):

    """
    Data structure representing input values from madx statements.

    These both an expression (str) and a value (bool/int/float).
    """

    def __init__(self, expr, value, type=float):
        """Store string expression and value."""
        self.expr = expr
        self._value = value
        self.type = type

    def __repr__(self):
        """Return string representation of this object."""
        return '{}({!r}, {}, {})'.format(self.__class__.__name__,
                                         self.expr, self.value,
                                         self.type.__name__)

    def __str__(self):
        """Get the expression as string."""
        return self.expr

    @property
    def value(self):
        """Get the value with the most accurate type."""
        return self.type(self._value)

    def __bool__(self):     # python3
        """Get the value as boolean."""
        return bool(self._value)

    __nonzero__ = __bool__   # python2

    def __int__(self):
        """Get the value as integer."""
        return int(self._value)

    def __float__(self):
        """Get the value as double."""
        return float(self._value)



"""
Installation script for CPyMAD.

Usage:
    python setup.py install --madxdir=/path/to/madx/installation

For more information, see
    http://hibtc.github.io/cpymad/installation
"""

# Make sure setuptools is available. NOTE: the try/except hack is required to
# make installation work with pip: If an older version of setuptools is
# already imported, `use_setuptools()` will just exit the current process.
try:
    import setuptools
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()

from setuptools import setup, Extension
from distutils.util import get_platform, convert_path
from distutils import sysconfig

import sys
from os import path


# setuptools.Extension automatically converts all '.pyx' extensions to '.c'
# extensions if detecting that neither Cython nor Pyrex is available. Early
# versions of setuptools don't know about Cython. Since we don't use Pyrex
# in this module, this leads to problems in the two cases where Cython is
# available and Pyrex is not or vice versa. Therefore, setuptools.Extension
# needs to be patched to match our needs:
try:
    # Use Cython if available:
    from Cython.Build import cythonize
except ImportError:
    # Otherwise, always use the distributed .c instead of the .pyx file:
    def cythonize(extensions):
        def pyx_to_c(source):
            return source[:-4]+'.c' if source.endswith('.pyx') else source
        for ext in extensions:
            ext.sources = list(map(pyx_to_c, ext.sources))
            missing_sources = [s for s in ext.sources if not path.exists(s)]
            if missing_sources:
                raise OSError(('Missing source file: {0[0]!r}. '
                               'Install Cython to resolve this problem.')
                              .format(missing_sources))
        return extensions
else:
    orig_Extension = Extension
    class Extension(orig_Extension):
        """Extension that *never* replaces '.pyx' by '.c' (using Cython)."""
        def __init__(self, name, sources, *args, **kwargs):
            orig_Extension.__init__(self, name, sources, *args, **kwargs)
            self.sources = sources


def fix_distutils_sysconfig_mingw():
    """
    When using windows and MinGW, in distutils.sysconfig the compiler (CC) is
    not initialized at all, see http://bugs.python.org/issue2437. The
    following manual fix for this problem may cause other issues, but it's a
    good shot.
    """
    if sysconfig.get_config_var('CC') is None:
        sysconfig._config_vars['CC'] = 'gcc'


def read_file(path):
    """Read a file in binary mode."""
    with open(convert_path(path), 'rb') as f:
        return f.read()


def exec_file(path):
    """Execute a python file and return the `globals` dictionary."""
    namespace = {}
    exec(read_file(path), namespace, namespace)
    return namespace


def get_long_description():
    """Compose a long description for PyPI."""
    long_description = None
    try:
        long_description = read_file('README.rst').decode('utf-8')
        long_description += '\n' + read_file('COPYING.rst').decode('utf-8')
        long_description += '\n' + read_file('CHANGES.rst').decode('utf-8')
    except (IOError, UnicodeDecodeError):
        pass
    return long_description


def get_extension_args(argv):
    """Get arguments for C-extension (include pathes, libraries, etc)."""
    # Let's just use the default system headers:
    include_dirs = []
    library_dirs = []
    # Parse command line option: --madxdir=/path/to/madxinstallation. We could
    # use build_ext.user_options instead, but then the --madxdir argument can
    # be passed only to the 'build_ext' command, not to 'build' or 'install',
    # which is a minor nuisance.
    for arg in argv:
        if arg.startswith('--madxdir='):
            argv.remove(arg)
            prefix = path.expanduser(arg.split('=', 1)[1])
            lib_path_candidates = [path.join(prefix, 'lib'),
                                path.join(prefix, 'lib64')]
            include_dirs += [path.join(prefix, 'include')]
            library_dirs += list(filter(path.isdir, lib_path_candidates))
    # required libraries
    if get_platform() == "win32" or get_platform() == "win-amd64":
        libraries = ['madx', 'stdc++', 'ptc', 'gfortran']
    else:
        libraries = ['madx', 'stdc++', 'c']
    # Common arguments for the Cython extensions:
    return dict(
        libraries=libraries,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        runtime_library_dirs=library_dirs,
        extra_compile_args=['-std=gnu99'],
    )


def get_setup_args(argv):
    """Accumulate metadata for setup."""
    extension_args = get_extension_args(argv)
    long_description = get_long_description()
    metadata = exec_file('cpymad/__init__.py')
    return dict(
        name='cpymad',
        version=metadata['__version__'],
        description=metadata['__summary__'],
        long_description=long_description,
        author=metadata['__author__'],
        author_email=metadata['__author_email__'],
        maintainer=metadata['__maintainer__'],
        maintainer_email=metadata['__maintainer_email__'],
        url=metadata['__uri__'],
        license=metadata['__license__'],
        classifiers=metadata['__classifiers__'],
        ext_modules = cythonize([
            Extension('cpymad.libmadx',
                    sources=["cpymad/libmadx.pyx"],
                    **extension_args),
        ]),
        packages = [
            "cpymad",
            "cpymad.resource",
        ],
        include_package_data=True, # include files matched by MANIFEST.in
        install_requires=[
            'setuptools',
            'numpy',
            'PyYAML',
        ],
    )


def main():
    """
    Execute setup.

    Note that this operation is controlled via sys.argv and has side-effects
    both on sys.argv as well as some parts of the distutils package.
    """
    fix_distutils_sysconfig_mingw()
    setup_args = get_setup_args(sys.argv)
    setup(**setup_args)


if __name__ == '__main__':
    main()



"""
Utility functions used in other parts of the pymad package.
"""
import collections
import re

from .types import Range, Constraint


__all__ = [
    'mad_quote',
    'is_identifier',
    'strip_element_suffix',
    'add_element_suffix',
    'mad_parameter',
    'mad_command',
]


try:
    unicode
except NameError:   # python3
    basestring = unicode = str


def mad_quote(value):
    """Add quotes to a string value."""
    quoted = repr(value)
    return quoted[1:] if quoted[0] == 'u' else quoted


# precompile regexes for performance:
_re_is_identifier = re.compile(r'^[a-z_]\w*$', re.IGNORECASE)
_re_element_suffix = re.compile(':\d+$')


def is_identifier(name):
    """Check if ``name`` is a valid identifier in MAD-X."""
    return bool(_re_is_identifier.match(name))


def strip_element_suffix(element_name):
    """
    Strip the :d suffix from an element name.

    The :d suffix is needed for some parts of the MAD-X API, but must not be
    used in other parts.
    """
    return _re_element_suffix.sub('', element_name)


def add_element_suffix(element_name):
    """
    Add a :1 suffix to an element name if missing.

    The :d suffix is needed for some parts of the MAD-X API, but must not be
    used in other parts.
    """
    if _re_element_suffix.search(element_name):
        return element_name
    return element_name + ':1'


def normalize_range_name(name):
    """Make element name usable as argument to the RANGE attribute."""
    if isinstance(name, tuple):
        return tuple(map(normalize_range_name, name))
    # MAD-X does not allow the ":d" suffix in the 'range' parameter string.
    # This means that name becomes less unique, but that's the only way right
    # now:
    name = strip_element_suffix(name)
    name = name.lower()
    if name.endswith('$end'):
        return '#e'
    if name.endswith('$start'):
        return '#s'
    return name


SOON = 1
SAFE = 10
LATE = 100
LAST = 1000


def mad_param_order(key, value):
    """
    Determine parameter order.

    Sadly, the parameter order actually matters - because of deficiencies in
    the MAD-X language and parser. For example the following cases can be
    problematic:

    - booleans after strings: "beam, sequence=s1, -radiate;"
    - after string lists: "select, flag=twiss, column=x,y, -full;"
    """
    key = str(key).lower()
    # the empty string was used in earlier versions in place of None:
    if value is None or value == '':
        return SAFE
    if isinstance(value, Range):
        return LATE
    elif isinstance(value, Constraint):
        return SAFE
    elif isinstance(value, bool):
        return SOON
    elif key == 'range':
        return LATE
    # check for basestrings before collections.Sequence, because every
    # basestring is also a Sequence:
    elif isinstance(value, basestring):
        return SAFE
    elif isinstance(value, collections.Sequence):
        if key == 'column':
            return LAST
        elif value and all(isinstance(v, basestring) for v in value):
            return LAST
        else:
            return SAFE
    else:
        return LATE


def mad_parameter(key, value):
    """
    Format a single MAD-X command parameter.
    """
    key = str(key).lower()
    # the empty string was used in earlier versions in place of None:
    if value is None or value == '':
        return ''
    if isinstance(value, Range):
        begin, end = normalize_range_name((value.first, value.last))
        return key + '=' + begin + '/' + end
    elif isinstance(value, Constraint):
        constr = []
        if value.min is not None:
            constr.append(key + '>' + value.min)
        if value.max is not None:
            constr.append(key + '<' + value.max)
        if constr:
            return ', '.join(constr)
        else:
            return key + '=' + value.value
    elif isinstance(value, bool):
        return ('' if value else '-') + key
    elif key == 'range':
        if isinstance(value, basestring):
            return key + '=' + normalize_range_name(value)
        elif isinstance(value, collections.Mapping):
            begin, end = value['first'], value['last']
        else:
            begin, end = value[0], value[1]
        begin, end = normalize_range_name((str(begin), str(end)))
        return key + '=' + begin + '/' + end
    # check for basestrings before collections.Sequence, because every
    # basestring is also a Sequence:
    elif isinstance(value, basestring):
        if key == 'file':
            return key + '=' + mad_quote(value)
        else:
            # MAD-X parses strings incorrectly, if followed by a boolean.
            # E.g.: "beam, sequence=s1, -radiate;" does NOT work! Therefore,
            # these values need to be quoted. (NOTE: MAD-X uses lower-case
            # internally and the quotes prevent automatic case conversion)
            return key + '=' + mad_quote(value.lower())
    elif isinstance(value, collections.Sequence):
        if key == 'column':
            return key + '=' + ','.join(value)
        elif value and all(isinstance(v, basestring) for v in value):
            return key + '=' + ','.join(value)
        else:
            return key + '={' + ','.join(map(str, value)) + '}'
    else:
        return key + '=' + str(value)


def mad_command(*args, **kwargs):
    """
    Create a MAD-X command from its name and parameter list.

    :param args: initial bareword command arguments (including command name!)
    :param kwargs: following named command arguments
    :returns: command string
    :rtype: str

    Examples:

    >>> mad_command('twiss', sequence='lhc')
    'twiss, sequence=lhc;'

    >>> mad_command('option', echo=True)
    'option, echo;'

    >>> mad_command('constraint', betx=Constraint(max=3.13))
    'constraint, betx<3.13;'
    """
    _args = list(args)
    _items = sorted(kwargs.items(), key=lambda v: mad_param_order(*v))
    _args += [mad_parameter(k, v) for k,v in _items]
    return ', '.join(filter(None, _args)) + ';'


def is_match_param(v):
    return v.lower() in ['rmatrix', 'chrom', 'beta0', 'deltap',
            'betx','alfx','mux','x','px','dx','dpx',
            'bety','alfy','muy','y','py','dy','dpy' ]



# encoding: utf-8
"""
Unit tests for the resource components
"""
# from __future__ import unicode_literals

# NOTE: The python2 setuptools.setup call expects pure binary (bytes) input
# parameters. It is therefore impractical to use module-level unicode
# support via the prededing __future__ import. The tests work on python2
# anyway since implicit coercion between unicode and bytes is possible. On
# python3 this is not possible, but not needed.

# tested classes
from cpymad.resource.package import PackageResource
from cpymad.resource.file import FileResource

# test utilities
import unittest
from yaml import safe_load as load
import os.path
import sys
import setuptools
import contextlib
from io import open

def set_(iterable):
    return set((s for s in iterable
                if not s.endswith('.pyc') and s != '__pycache__'))


@contextlib.contextmanager
def captured_output(stream_name):
    if str is bytes:
        # Python2: setuptools gives non-unicode output, so we have to
        # supply a binary stream:
        from io import BytesIO as StringIO
    else:
        # Python3: str is unicode and everything is alright.
        from io import StringIO
    orig_stdout = getattr(sys, stream_name)
    setattr(sys, stream_name, StringIO())
    try:
        yield getattr(sys, stream_name)
    finally:
        setattr(sys, stream_name, orig_stdout)


# common test code

class Common(object):

    def setUp(self):
        here = os.path.dirname(__file__)
        self.base = os.path.join(here, 'data')
        self.mod = 'package'
        self.path = os.path.join(self.base, self.mod)

    def tearDown(self):
        pass

    def test_open(self):
        with self.res.open('a.yml', 'utf-8') as f:
            self.assertEqual(
                load(f.read())['path'],
                'a.yml')
        with self.res.get('subdir/b.yml').open(encoding='utf-8') as f:
            self.assertEqual(
                    load(f.read())['path'],
                    os.path.join('subdir', 'b.yml'))

    def test_list(self):
        self.assertEqual(
                set_(self.res.listdir()),
                set(('__init__.py', 'a.yml', 'subdir')))
        self.assertEqual(
                set(self.res.listdir('subdir')),
                set(('b.yml',)))
        self.assertEqual(
                set(self.res.get('subdir').listdir()),
                set(('b.yml',)))

    def test_provider(self):
        self.assertEqual(
                set_(self.res.get('subdir').provider().listdir()),
                set(('__init__.py', 'a.yml', 'subdir')))
        self.assertEqual(
                set(self.res.get(['subdir', 'b.yml']).provider().listdir()),
                set(('b.yml',)))

    def test_filter(self):
        self.assertEqual(
                set(self.res.listdir_filter(ext='.txt')),
                set())
        self.assertEqual(
                set(self.res.listdir_filter(ext='.yml')),
                set(('a.yml',)))

    def test_load(self):
        self.assertEqual(
                load(self.res.load('a.yml', 'utf-8'))['path'],
                'a.yml')
        self.assertEqual(
                load(self.res.get('subdir').load('b.yml', 'utf-8'))['path'],
                os.path.join('subdir', 'b.yml'))

    def test_yaml(self):
        self.assertEqual(
                self.res.yaml('a.yml')['path'],
                'a.yml')
        self.assertEqual(
                self.res.get(['subdir', 'b.yml']).yaml()['path'],
                os.path.join('subdir', 'b.yml'))


# test cases
class TestPackageResource(Common, unittest.TestCase):

    def setUp(self):
        super(TestPackageResource, self).setUp()
        sys.path.append(self.base)
        self.res = PackageResource(self.mod)

    def tearDown(self):
        del self.res
        del sys.modules[self.mod]
        sys.path.remove(self.base)
        super(TestPackageResource, self).tearDown()

    def test_filename(self):
        with self.res.filename('a.yml') as filename:
            with open(filename, encoding='utf-8') as f:
                self.assertEqual(
                        load(f.read())['path'],
                        'a.yml')
        self.assertTrue(os.path.exists(filename))
        with self.res.get(['subdir', 'b.yml']).filename() as filename:
            with open(filename, encoding='utf-8') as f:
                self.assertEqual(
                        load(f.read())['path'],
                        'subdir/b.yml')
        self.assertTrue(os.path.exists(filename))

class TestEggResource(Common, unittest.TestCase):

    def setUp(self):
        super(TestEggResource, self).setUp()
        cwd = os.getcwd()
        os.chdir(self.base)
        with captured_output('stdout'):
            with captured_output('stderr'):
                setuptools.setup(
                    name=self.mod,
                    packages=[self.mod],
                    script_args=['bdist_egg', '--quiet'],
                    package_data={self.mod:[
                        'a.yml',
                        os.path.join('subdir', 'b.yml')]})
        os.chdir(cwd)
        self.eggs = os.listdir(os.path.join(self.base, 'dist'))
        for egg in self.eggs:
            sys.path.append(os.path.join(self.base, 'dist', egg))
        self.res = PackageResource(self.mod)

    def tearDown(self):
        del self.res
        del sys.modules[self.mod]
        for egg in self.eggs:
            sys.path.remove(os.path.join(self.base, 'dist', egg))
        super(TestEggResource, self).tearDown()

    def test_is_valid(self):
        """
        Check that the test uses a .egg resource.

        Only then the test results confirm that the package is working.

        """
        self.assertTrue(self.res._is_extracted)

    def test_filename(self):
        with self.res.filename('a.yml') as filename:
            with open(filename, encoding='utf-8') as f:
                self.assertEqual(
                        load(f.read())['path'],
                        'a.yml')
        self.assertFalse(os.path.exists(filename))
        with self.res.get(['subdir', 'b.yml']).filename() as filename:
            with open(filename, encoding='utf-8') as f:
                self.assertEqual(
                        load(f.read())['path'],
                        'subdir/b.yml')
        self.assertFalse(os.path.exists(filename))
        # The resource should be accessible again, even though the file was
        # deleted when exiting the context manager:
        with self.res.get(['subdir', 'b.yml']).filename() as filename:
            with open(filename, encoding='utf-8') as f:
                self.assertEqual(
                        load(f.read())['path'],
                        'subdir/b.yml')
        self.assertFalse(os.path.exists(filename))

    def test_use_filename_twice(self):
        with self.res.filename('a.yml') as filename:
            self.assertTrue(os.path.exists(filename))
        self.assertFalse(os.path.exists(filename))
        with self.res.filename('a.yml') as filename:
            self.assertTrue(os.path.exists(filename))
            with open(filename, encoding='utf-8') as f:
                self.assertEqual(
                        load(f.read())['path'],
                        'a.yml')

class TestFileResource(Common, unittest.TestCase):

    def setUp(self):
        super(TestFileResource, self).setUp()
        self.res = FileResource(self.path)

    def tearDown(self):
        del self.res
        super(TestFileResource, self).tearDown()

    def test_filename(self):
        with self.res.filename('a.yml') as filename:
            self.assertEqual(
                    filename,
                    os.path.join(self.path, 'a.yml'))
        with self.res.get(['subdir', 'b.yml']).filename() as filename:
            self.assertEqual(
                    filename,
                    os.path.join(self.path, 'subdir', 'b.yml'))


if __name__ == '__main__':
    unittest.main()
